import os
from typing import List

from requests_futures.sessions import FuturesSession

from savvihub.api.file_object import UploadableFileObject, DownloadableFileObject
from savvihub.api.savvihub import SavviHubClient
from savvihub.api.types import SavviHubFileObject
from savvihub.common.utils import wait_all_futures, calculate_crc32c


def default_hooks():
    def fn(resp, **kwargs):
        resp.raise_for_status()
    return {
        'response': fn,
    }


class Downloader:
    @classmethod
    def download(cls, local_path, file: SavviHubFileObject, progressable=None):
        dirname = os.path.dirname(local_path)
        basename = os.path.basename(local_path)
        d = DownloadableFileObject(file.download_url.get('url'), dirname, basename)
        d.download(progressable=progressable)

    @classmethod
    def bulk_download(cls, context, local_base_path, remote_files: List[SavviHubFileObject], progressable=None):
        if len(remote_files) <= 0:
            return
        session = FuturesSession(max_workers=context.parallel)
        futures = []
        for remote_file in remote_files:
            if remote_file.path.endswith('/'):
                continue

            d = DownloadableFileObject(remote_file.download_url.get('url'), local_base_path, remote_file.path)
            futures.append(d.download(session, progressable=progressable))

        wait_all_futures(futures)


class Uploader:
    @classmethod
    def get_files_to_upload(cls, local_base_path, hashmap=None):
        local_base_path = local_base_path.rstrip('/')
        results = []
        for root, dirs, files in os.walk(local_base_path):
            for name in files:
                name = os.path.join(os.path.abspath(root), name)
                name = name[len(local_base_path) + 1:] if name.startswith(local_base_path) else name
                if hashmap and hashmap[name] == calculate_crc32c(os.path.join(local_base_path, name)):
                    continue
                results.append(name)
        return results

    @classmethod
    def get_hashmap(cls, local_base_path):
        files = cls.get_files_to_upload(local_base_path)
        hashmap = dict()
        for file in files:
            path = os.path.join(local_base_path, file)
            hashmap[file] = calculate_crc32c(path)
        return hashmap

    @classmethod
    def upload(cls, context, local_path, volume_id, remote_path, progressable=None):
        client = SavviHubClient(token=context.token)
        file_create_resp = client.volume_file_create(volume_id, remote_path, is_dir=False, raise_error=True)
        file_object = SavviHubFileObject(file_create_resp.json())

        dirname = os.path.dirname(local_path)
        basename = os.path.basename(local_path)
        u = UploadableFileObject(file_object.upload_url.get("url"), dirname, basename)
        u.upload(progressable=progressable)
        resp = client.volume_file_uploaded(volume_id, remote_path, raise_error=True)
        return SavviHubFileObject(resp.json())

    @classmethod
    def bulk_upload(cls, context, local_base_path, local_file_paths, volume_id, remote_base_path, *, progressable=None):
        if len(local_file_paths) <= 0:
            return

        session = FuturesSession(max_workers=context.parallel)
        client = SavviHubClient(token=context.token, session=session)

        futures = []
        for local_file_path in local_file_paths:
            future = client.volume_file_create(
                volume_id,
                os.path.join(remote_base_path, local_file_path),
                is_dir=False,
                hooks=default_hooks()
            )
            futures.append(future)
        resps = wait_all_futures(futures)

        futures = []
        for i, resp in enumerate(resps):
            file_object = SavviHubFileObject(resp.json())

            u = UploadableFileObject(file_object.upload_url.get("url"), local_base_path, local_file_paths[i])
            futures.append(u.upload(session, progressable=progressable))

        wait_all_futures(futures)

        futures = []
        for local_file_path in local_file_paths:
            future = client.volume_file_uploaded(
                volume_id,
                os.path.join(remote_base_path, local_file_path),
                hooks=default_hooks(),
            )
            futures.append(future)

        resps = wait_all_futures(futures)
        return [SavviHubFileObject(resp.json()) for resp in resps]



import re
from unittest.mock import patch

from savvihub.common.constants import WEB_HOST
from savvihub.common.utils import random_string
from tests.conftest import BaseTestCase


class ExperimentTest(BaseTestCase):
    def test_experiment(self):
        result = self.runner.invoke(['experiment', 'list'])
        assert f'There is no experiment in {self.project_name} project.' in result.output

        # Create a dataset
        dataset_name = random_string()
        result = self.runner.invoke(['dataset', 'create', f'{self.workspace_name}/{dataset_name}'])
        assert f'Dataset {dataset_name} is created.' in result.output
        assert f'{WEB_HOST}/{self.workspace_name}/datasets/{dataset_name}' in result.output

        # Run a new experiment with --git-ref
        result = self.runner.invoke(['experiment', 'run',
                                     '-c', 'python main.py',
                                     '-i', 'savvihub/kernels:full-cpu',
                                     '-r', 'resource-spec',
                                     '--git-ref', 'dummy-git-ref',
                                     '--dataset', f'{self.workspace_name}/{dataset_name}:/input'])
        assert f'{WEB_HOST}/{self.workspace_name}/{self.project_name}/experiments/1' in result.output

        result = self.runner.invoke(['experiment', 'describe', '1'])
        experiment_name = re.search('^Name: (.+)$', result.output, re.M).group(1)

        result = self.runner.invoke(['experiment', 'list'])
        assert experiment_name in result.output

        result = self.runner.invoke(['experiment', 'logs', '1'])
        assert f'{WEB_HOST}/{self.workspace_name}/{self.project_name}/experiments/1/logs' in result.output

        result = self.runner.invoke(['experiment', 'output', 'ls', '1'])
        assert 'This directory is empty.' in result.output
        result = self.runner.invoke(['experiment', 'output', 'download', '1'])
        assert 'Found 0 files to download.' in result.output

        # Run a new experiment without --git-ref (git diff created)
        with patch('inquirer.prompt', return_value={'experiment': '[1] Run experiment with uncommitted and untracked changes.'}):
            # Run a new experiment without --git-ref
            result = self.runner.invoke(['experiment', 'run',
                                         '-c', 'python main.py',
                                         '-i', 'savvihub/kernels:full-cpu',
                                         '-r', 'resource-spec',
                                         '--dataset', f'{self.workspace_name}/{dataset_name}:/input'])
            assert 'Generating diff patch file...' in result.output
            assert f'{WEB_HOST}/{self.workspace_name}/{self.project_name}/experiments/2' in result.output
            result = self.runner.invoke(['experiment', 'describe', '2'])
            assert 'Git Diff File:' in result.output
            git_diff_file_url = re.search('URL: (.+)$', result.output, re.M).group(1)

        # Run a new experiment with --git-ref and --git-diff
        result = self.runner.invoke(['experiment', 'run',
                                     '-c', 'python main.py',
                                     '-i', 'savvihub/kernels:full-cpu',
                                     '-r', 'resource-spec',
                                     '--git-ref', 'dummy-git-ref',
                                     '--git-diff', git_diff_file_url,
                                     '--dataset', f'{self.workspace_name}/{dataset_name}:/input'])
        assert 'Generating diff patch file...' in result.output
        assert f'{WEB_HOST}/{self.workspace_name}/{self.project_name}/experiments/3' in result.output
        result = self.runner.invoke(['experiment', 'describe', '3'])
        assert 'Git Diff File:' in result.output

        # Run a new experiment without --git-ref but do not use diff
        with patch('inquirer.prompt', return_value={'experiment': '[3] Run experiment without any changes.'}):
            # Run a new experiment without --git-ref
            result = self.runner.invoke(['experiment', 'run',
                                         '-c', 'python main.py',
                                         '-i', 'savvihub/kernels:full-cpu',
                                         '-r', 'resource-spec',
                                         '--dataset', f'{self.workspace_name}/{dataset_name}:/input'])
            assert 'Generating diff patch file...' not in result.output
            assert f'{WEB_HOST}/{self.workspace_name}/{self.project_name}/experiments/4' in result.output
            result = self.runner.invoke(['experiment', 'describe', '4'])
            assert 'Git Diff File:' not in result.output

        # Try to run a new experiment without --git-ref and abort
        with patch('inquirer.prompt', return_value={'experiment': '[4] Abort.'}):
            # Run a new experiment without --git-ref
            result = self.runner.invoke(['experiment', 'run',
                                         '-c', 'python main.py',
                                         '-i', 'savvihub/kernels:full-cpu',
                                         '-r', 'resource-spec',
                                         '--dataset', f'{self.workspace_name}/{dataset_name}:/input'])
            assert 'Aborted.' in result.output



import tempfile
from typing import Optional, Union, Iterable, Text, IO, Any, Mapping
from unittest import TestCase

from click.testing import Result
from typer.testing import CliRunner as TyperCliRunner

from savvihub import Context
from savvihub.api.savvihub import SavviHubClient
from savvihub.common import constants
from savvihub.common.utils import random_string
from savvihub.savvihub_cli.main import app


class CliRunner(TyperCliRunner):
    def invoke(
        self,
        args: Optional[Union[str, Iterable[str]]] = None,
        input: Optional[Union[bytes, Text, IO[Any]]] = None,
        env: Optional[Mapping[str, str]] = None,
        color: bool = False,
        allow_fail: bool = False,
        **extra: Any,
    ) -> Result:
        result = super().invoke(app, args, input, env, catch_exceptions=False, color=color, **extra)
        if not allow_fail:
            assert result.exit_code == 0, result.output
        return result


class BaseTestCase(TestCase):
    def setUp(self):
        super().setUp()
        constants.TEST = True
        self._auto_savvi_init()
        self._monkey_patch(self.token, self.workspace_name, self.project_name)
        self.context = Context(auth_required=True, project_required=True)
        self.client = SavviHubClient(token=self.context.token)
        self.runner = CliRunner()

    def _auto_savvi_init(self):
        # user init
        username = random_string()
        jwt_client = SavviHubClient()
        r = jwt_client.post('/api/v1/accounts/signup', {
            'email': f'{username}@savvihub.test',
            'username': username,
            'name': username,
            'password': 'testtest',
            'invitation_token': 'invitation_token_for_cli_test',
        }, raise_error=True).json()
        jwt_client.session.headers['Authorization'] = f'JWT {r["token"]}'

        cli_token = jwt_client.check_signin_token()
        jwt_client.post('/api/v1/accounts/signin/cli/confirm', {
            'cli_token': cli_token,
        }, raise_error=True)

        success, access_token = jwt_client.check_signin(cli_token)
        if not success:
            raise Exception('Signin Failed')

        # project init
        access_token_client = SavviHubClient(token=access_token)
        workspace_name = access_token_client.workspace_list()[0].name
        project_name = random_string()
        me = access_token_client.get_my_info()
        access_token_client.project_github_create(workspace_name, project_name, me.username, project_name)

        self.token = access_token
        self.workspace_name = workspace_name
        self.project_name = project_name

    @staticmethod
    def _monkey_patch(token, workspace_name, project_name):
        class MockConfig:
            def __init__(self):
                self.token = token
                self.workspace = workspace_name
                self.project = project_name

        class MockGitRepo:
            def check_revision_in_remote(self, ref):
                return True

            def get_remote_revision_or_branch(self):
                return 'dummy-ref', 'dummy-branch', False

            def get_commit_message(self, ref):
                return 'dummy-commit-message'

            def get_current_diff_status(self, ref):
                return True, {
                    'untracked': ['dummy-untracked-file.py', 'dummy-untracked-file.go'],
                    'uncommitted': ['dummy-uncommitted-file.py', 'dummy-uncommitted-file.go'],
                }

            def get_current_diff_file(self, git_ref, with_untracked=True):
                fp = tempfile.NamedTemporaryFile(suffix='.patch')
                fp.write('dummy-diff'.encode())
                return fp

        def context_init(
            self,
            auth_required=False,
            user_required=False,
            project_required=False,
            experiment_required=False,
        ):
            mock_config = MockConfig()
            self.global_config = mock_config

            if user_required:
                self.user = self.get_my_info()

            if project_required:
                self.git_repo = MockGitRepo()
                self.project_config = mock_config
                self.project = self.get_project()

            if experiment_required:
                raise NotImplementedError('Please implement this..')

        Context.__init__ = context_init



import os
from pathlib import Path


DEBUG = os.environ.get('SAVVIHUB_DEBUG', False)
API_HOST = os.environ.get('SAVVIHUB_API_HOST', 'https://api.savvihub.com')
WEB_HOST = os.environ.get('SAVVIHUB_WEB_HOST', 'https://savvihub.com')

CUR_DIR = os.getcwd()
DEFAULT_SAVVI_DIR = os.path.join(str(Path.home()), '.savvihub')
DEFAULT_CONFIG_PATH = os.path.join(DEFAULT_SAVVI_DIR, 'config')
PROJECT_ROOT = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

INQUIRER_NAME_IMAGE = 'image'
INQUIRER_NAME_RESOURCE = 'resource'
INQUIRER_NAME_DATASET = 'dataset'
INQUIRER_NAME_DATASET_REF = 'dataset_ref'
INQUIRER_NAME_DATASET_MOUNT_PATH = 'dataset_mount_path'
INQUIRER_NAME_COMMAND = 'command'
INQUIRER_NAME_CPU_LIMIT = 'cpu_limit'
INQUIRER_NAME_GPU_LIMIT = 'gpu_limit'
INQUIRER_NAME_MEMORY_LIMIT = 'memory_limit'

ROLETYPE_DATASET_FILES = 'dataset-files'
ROLETYPE_EXPERIMENT_OUTPUT = 'experiment-output'
ROLETYPE_EXPERIMENT_INPUT = 'experiment-input'

DATASET_SOURCE_TYPE_SAVVIHUB = 'savvihub'

DATASET_PATH_PARSE_SCHEME_GS = "gs"
DATASET_PATH_PARSE_SCHEME_S3 = "S3"



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


import inspect
import pprint
import re  # noqa: F401
import six

from savvihub_client.configuration import Configuration


class ResponseVolumeSourceDataset(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'dataset': 'ResponseDataset',
        'dataset_id': 'int',
        'snapshot': 'ResponseSnapshot',
        'snapshot_id': 'int',
        'sub_path': 'str'
    }

    attribute_map = {
        'dataset': 'dataset',
        'dataset_id': 'dataset_id',
        'snapshot': 'snapshot',
        'snapshot_id': 'snapshot_id',
        'sub_path': 'sub_path'
    }

    def __init__(self, dataset=None, dataset_id=None, snapshot=None, snapshot_id=None, sub_path=None, local_vars_configuration=None):  # noqa: E501
        """ResponseVolumeSourceDataset - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._dataset = None
        self._dataset_id = None
        self._snapshot = None
        self._snapshot_id = None
        self._sub_path = None
        self.discriminator = None

        if dataset is not None:
            self.dataset = dataset
        if dataset_id is not None:
            self.dataset_id = dataset_id
        if snapshot is not None:
            self.snapshot = snapshot
        if snapshot_id is not None:
            self.snapshot_id = snapshot_id
        if sub_path is not None:
            self.sub_path = sub_path

    @property
    def dataset(self):
        """Gets the dataset of this ResponseVolumeSourceDataset.  # noqa: E501


        :return: The dataset of this ResponseVolumeSourceDataset.  # noqa: E501
        :rtype: ResponseDataset
        """
        return self._dataset

    @dataset.setter
    def dataset(self, dataset):
        """Sets the dataset of this ResponseVolumeSourceDataset.


        :param dataset: The dataset of this ResponseVolumeSourceDataset.  # noqa: E501
        :type dataset: ResponseDataset
        """

        self._dataset = dataset

    @property
    def dataset_id(self):
        """Gets the dataset_id of this ResponseVolumeSourceDataset.  # noqa: E501


        :return: The dataset_id of this ResponseVolumeSourceDataset.  # noqa: E501
        :rtype: int
        """
        return self._dataset_id

    @dataset_id.setter
    def dataset_id(self, dataset_id):
        """Sets the dataset_id of this ResponseVolumeSourceDataset.


        :param dataset_id: The dataset_id of this ResponseVolumeSourceDataset.  # noqa: E501
        :type dataset_id: int
        """

        self._dataset_id = dataset_id

    @property
    def snapshot(self):
        """Gets the snapshot of this ResponseVolumeSourceDataset.  # noqa: E501


        :return: The snapshot of this ResponseVolumeSourceDataset.  # noqa: E501
        :rtype: ResponseSnapshot
        """
        return self._snapshot

    @snapshot.setter
    def snapshot(self, snapshot):
        """Sets the snapshot of this ResponseVolumeSourceDataset.


        :param snapshot: The snapshot of this ResponseVolumeSourceDataset.  # noqa: E501
        :type snapshot: ResponseSnapshot
        """

        self._snapshot = snapshot

    @property
    def snapshot_id(self):
        """Gets the snapshot_id of this ResponseVolumeSourceDataset.  # noqa: E501


        :return: The snapshot_id of this ResponseVolumeSourceDataset.  # noqa: E501
        :rtype: int
        """
        return self._snapshot_id

    @snapshot_id.setter
    def snapshot_id(self, snapshot_id):
        """Sets the snapshot_id of this ResponseVolumeSourceDataset.


        :param snapshot_id: The snapshot_id of this ResponseVolumeSourceDataset.  # noqa: E501
        :type snapshot_id: int
        """

        self._snapshot_id = snapshot_id

    @property
    def sub_path(self):
        """Gets the sub_path of this ResponseVolumeSourceDataset.  # noqa: E501


        :return: The sub_path of this ResponseVolumeSourceDataset.  # noqa: E501
        :rtype: str
        """
        return self._sub_path

    @sub_path.setter
    def sub_path(self, sub_path):
        """Sets the sub_path of this ResponseVolumeSourceDataset.


        :param sub_path: The sub_path of this ResponseVolumeSourceDataset.  # noqa: E501
        :type sub_path: str
        """

        self._sub_path = sub_path

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = inspect.getargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ResponseVolumeSourceDataset):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ResponseVolumeSourceDataset):
            return True

        return self.to_dict() != other.to_dict()



import json
import sys


class SavviHubError(Exception):
    error_code = 'UnexpectedProblem'
    _message = 'Unexpected problem found.'
    fields = []

    def __init__(self, message, fields=None):
        if message:
            self._message = message
        if fields:
            self.fields = fields

    def __eq__(self, other):
        return self.error_code == other.error_code

    def message(self):
        msg = self._message
        if self.fields:
            msg += ' ('
            for i, field in enumerate(self.fields):
                if i > 0:
                    msg += ', '
                msg += field['name']
                if field['value']:
                    msg += f': {field["value"]}'
            msg += ')'

        return msg


def inheritors(klass):
    subclasses = []
    work = [klass]
    while work:
        parent = work.pop()
        for child in parent.__subclasses__():
            if child not in subclasses:
                subclasses.append(child)
                work.append(child)
    return subclasses


class SavviHubUnexpectedProblemError(SavviHubError):
    error_code = 'UnexpectedProblem'


class SavviHubNotADirectoryError(SavviHubError):
    error_code = 'NotADirectory'


class SavviHubNoSuchFileOrDirectoryError(SavviHubError):
    error_code = 'NoSuchFileOrDirectory'


class SavviHubK8sUnauthorizedError(SavviHubError):
    error_code = 'K8sUnauthorized'


class SavviHubK8sPermissionDeniedError(SavviHubError):
    error_code = 'K8sPermissionDenied'


class SavviHubK8sTimeoutError(SavviHubError):
    error_code = 'K8sTimeout'


class SavviHubConnectionRefusedError(SavviHubError):
    error_code = 'ConnectionRefused'


class SavviHubInvalidCACertError(SavviHubError):
    error_code = 'InvalidCACert'


class SavviHubNetworkTimeoutError(SavviHubError):
    error_code = 'NetworkTimeout'


def parse_error_code(resp):
    try:
        code = resp.json().get('code')
        message = resp.json().get('message')
        fields = resp.json().get('fields')

        for klass in inheritors(SavviHubError):
            if klass.error_code == code:
                return klass(message, fields)

        return SavviHubUnexpectedProblemError(message, fields)

    except Exception as e:
        return SavviHubUnexpectedProblemError(sys.exc_info()[0])



import sys
import typing

import requests
import typer

from savvihub.api.errors import parse_error_code, SavviHubError, SavviHubNotADirectoryError, \
    SavviHubNoSuchFileOrDirectoryError
from savvihub.api.types import PaginatedMixin, SavviHubListResponse, SavviHubFileObject, SavviHubDataset, \
    SavviHubKernelResource, SavviHubKernelImage, SavviHubExperiment, SavviHubProject, SavviHubWorkspace, SavviHubUser, \
    SavviHubExperimentLog, SavviHubSnapshot, SavviHubVolume, SavviHubExperimentMetric, SavviHubKernelCluster
from savvihub.common.constants import API_HOST, WEB_HOST, DEBUG


class SavviHubClient:
    def __init__(self, *, session=requests.Session(), token=None, url=API_HOST, content_type='application/json'):
        self.session = session
        self.url = url
        self.token = token
        
        session.headers = {'content-type': content_type}
        if token:
            session.headers['authorization'] = 'Token %s' % token

    def get(self, url, params=None, raise_error=False, **kwargs):
        r = self.session.get(f'{self.url}{url}', params=params, **kwargs)
        if DEBUG:
            print({
                'method': 'GET',
                'url': url,
                'params': params,
                'status_code': r.status_code if isinstance(r, requests.Response) else None,
                'response': r.text if isinstance(r, requests.Response) else None,
            })
        if raise_error and r.status_code / 100 >= 4:
            raise parse_error_code(r)
        return r

    def get_all(self, url, params=None, raise_error=False, **kwargs):
        raw_r = self.get(url, params=params, raise_error=raise_error, **kwargs)
        r = PaginatedMixin(raw_r.json())
        results = []

        fetched_items = 0
        while True:
            fetched_items += len(r.results)
            results.extend(r.results)
            if fetched_items >= r.total:
                break
            raw_r = self.get(url, params={**params, 'after': r.endCursor}, raise_error=raise_error, **kwargs)
            r = PaginatedMixin(raw_r.json())
        return results

    def get_all_without_pagination(self, url, params=None, raise_error=False, **kwargs):
        raw_r = self.get(url, params=params, raise_error=raise_error, **kwargs)
        r = SavviHubListResponse(raw_r.json())
        return r.results

    def post(self, url, data, raise_error=False, **kwargs):
        r = self.session.post(f'{self.url}{url}', json=data, **kwargs)
        if DEBUG:
            print({
                'method': 'POST',
                'url': url,
                'request': data,
                'status_code': r.status_code if isinstance(r, requests.Response) else None,
                'response': r.text if isinstance(r, requests.Response) else None,
            })
        if raise_error and r.status_code / 100 >= 4:
            raise parse_error_code(r)
        return r

    def delete(self, url, raise_error=False, **kwargs):
        r = self.session.delete(f'{self.url}{url}', **kwargs)
        if DEBUG:
            print({
                'method': 'DELETE',
                'url': url,
                'status_code': r.status_code if isinstance(r, requests.Response) else None,
                'response': r.text if isinstance(r, requests.Response) else None,
            })
        if raise_error and r.status_code / 100 >= 4:
            raise parse_error_code(r)
        return r

    def patch(self, url, data, raise_error=False, **kwargs):
        r = self.session.patch(f'{self.url}{url}', json=data, **kwargs)
        if DEBUG:
            print({
                'method': 'PATCH',
                'url': url,
                'request': data,
                'status_code': r.status_code if isinstance(r, requests.Response) else None,
                'response': r.text if isinstance(r, requests.Response) else None,
            })
        if raise_error and r.status_code / 100 >= 4:
            raise parse_error_code(r)
        return r

    def get_my_info(self):
        r = self.get(f'/api/v1/accounts/me')
        if r.status_code != 200:
            return None
        return SavviHubUser(r.json())

    def verify_access_token(self):
        r = self.post(f'/api/v1/accounts/access_tokens/verify', {})
        r.raise_for_status()

    def check_signin_token(self):
        r = self.post(f'/api/v1/accounts/signin/cli/token', {})
        r.raise_for_status()
        return r.json()['cli_token']

    def check_signin(self, cli_token):
        r = self.get(f'/api/v1/accounts/signin/cli/check', {
            'cli_token': cli_token,
        })
        r.raise_for_status()
        return r.json()['signin_success'], r.json()['access_token']

    def volume_read(self, volume_id, **kwargs):
        r = self.get(f'/api/v1/volumes/{volume_id}', **kwargs)
        return SavviHubVolume(r.json())

    def volume_file_list(
        self, volume_id, snapshot='latest', path='', recursive=False, need_download_url=False, **kwargs,
    ) -> typing.Optional[typing.List[SavviHubFileObject]]:
        try:
            r = self.get(f'/api/v1/volumes/{volume_id}/files', params={
                'path': path,
                'recursive': recursive,
                'snapshot': snapshot,
                'need_download_url': need_download_url,
            }, raise_error=True, **kwargs)
        except SavviHubError as e:
            if e == SavviHubNotADirectoryError:
                typer.echo(f'Not a directory: {path}')
            elif e == SavviHubNoSuchFileOrDirectoryError:
                typer.echo(f'No such file or directory: {path}')
            sys.exit(1)

        return [SavviHubFileObject(x) for x in r.json().get('results')]

    def volume_file_read(self, volume_id, path, snapshot, **kwargs):
        try:
            r = self.get(f'/api/v1/volumes/{volume_id}/files/read', params={
                'path': path, 'snapshot': snapshot,
            }, raise_error=True, **kwargs)
        except SavviHubError as e:
            if e == SavviHubNotADirectoryError:
                typer.echo(f'Not a directory: {path}')
            elif e == SavviHubNoSuchFileOrDirectoryError:
                return None
            print(e)
            sys.exit(1)

        return SavviHubFileObject(r.json())

    def volume_file_copy(self, volume_id, source_path, source_snapshot, dest_path, recursive=False, **kwargs):
        r = self.post(f'/api/v1/volumes/{volume_id}/files/copy', {
            'source_path': source_path,
            'source_snapshot': source_snapshot,
            'dest_path': dest_path,
            'recursive': recursive,
        }, **kwargs)
        r.raise_for_status()
        return [SavviHubFileObject(x) for x in r.json().get('copied_files')]

    def volume_file_delete(self, volume_id, path, recursive=False):
        r = self.delete(f'/api/v1/volumes/{volume_id}/files/delete', params={'path': path, 'recursive': recursive})
        if r.status_code != 200:
            return None
        return [SavviHubFileObject(x) for x in r.json().get('DeletedFiles')]

    def volume_file_create(self, volume_id, path, is_dir, **kwargs):
        return self.post(f'/api/v1/volumes/{volume_id}/files', {
            'path': path,
            'is_dir': is_dir
        }, **kwargs)

    def volume_file_uploaded(self, volume_id, path, **kwargs):
        return self.post(f'/api/v1/volumes/{volume_id}/files/uploaded?path={path}', {
        }, **kwargs)

    def snapshot_read(self, volume_id, ref, **kwargs):
        r = self.get(f'/api/v1/volumes/{volume_id}/snapshots/{ref}', **kwargs)
        if r.status_code != 200:
            return None
        return SavviHubSnapshot(r.json())

    def experiment_id_read(self, experiment_id, **kwargs):
        r = self.get(f'/api/v1/experiments/{experiment_id}', **kwargs)
        if r.status_code != 200:
            return None
        return SavviHubExperiment(r.json())

    def experiment_read(self, workspace, project, experiment_number_or_name, **kwargs) -> SavviHubExperiment:
        return SavviHubExperiment(
            self.get(f'/api/v1/workspaces/{workspace}/projects/{project}/experiments/{experiment_number_or_name}',
                     **kwargs).json())

    def experiment_list(self, workspace, project, **kwargs):
        return [SavviHubExperiment(x)
                for x in self.get_all_without_pagination(
                f'/api/v1/workspaces/{workspace}/projects/{project}/experiments?orderby.field=number&orderby.direction=desc', **kwargs)]

    def experiment_log(self, workspace, project, experiment_number_or_name, **kwargs):
        return [SavviHubExperimentLog(x)
                for x in self.get(
                f'/api/v1/workspaces/{workspace}/projects/{project}/experiments/{experiment_number_or_name}/log',
                **kwargs).json().get('logs')]

    def experiment_create(self, workspace, project, cluster_name, image_url, resource_spec_id, resource_spec,
                          git_ref, git_diff_file_path, start_command, datasets, env_vars, **kwargs):
        return self.post(f'/api/v1/workspaces/{workspace}/projects/{project}/experiments', {
            'cluster_name': cluster_name,
            'image_url': image_url,
            'resource_spec_id': resource_spec_id,
            'resource_spec': resource_spec,
            'git_ref': git_ref,
            'datasets': datasets,
            'git_diff_file_path': git_diff_file_path,
            'env_vars': env_vars,
            'start_command': start_command,
        }, **kwargs)

    def experiment_metrics_update(self, experiment_id, metrics: typing.Dict[str, typing.List[SavviHubExperimentMetric]], **kwargs):
        return self.post(f'/api/v1/experiments/{experiment_id}/metrics', {
            'metrics': metrics,
        }, **kwargs)

    def experiment_system_metrics_update(self, experiment_id, metrics: typing.Dict[str, typing.List[SavviHubExperimentMetric]], **kwargs):
        return self.post(f'/api/v1/experiments/{experiment_id}/system_metrics', {
            'system_metrics': metrics,
        }, **kwargs)

    def kernel_image_list(self, workspace, type):
        results = self.get_all_without_pagination(f'/api/v1/workspaces/{workspace}/kernels/images')
        return [SavviHubKernelImage(x) for x in results if x.get("processor_type") == type]

    def kernel_resource_list(self, workspace):
        results = self.get_all_without_pagination(f'/api/v1/workspaces/{workspace}/kernels/resource_specs')
        return [SavviHubKernelResource(x) for x in results]

    def workspace_list(self, **kwargs):
        r = self.get(f'/api/v1/workspaces', **kwargs)
        return [SavviHubWorkspace(x) for x in r.json().get('workspaces')]

    def workspace_read(self, workspace):
        r = self.get(f'/api/v1/workspaces/{workspace}')
        if r.status_code == 404:
            return None
        elif r.status_code != 200:
            r.raise_for_status()
        return SavviHubWorkspace(r.json())

    def project_read(self, workspace, project):
        r = self.get(f'/api/v1/workspaces/{workspace}/projects/{project}')
        if r.status_code == 404:
            return None
        elif r.status_code != 200:
            r.raise_for_status()
        return SavviHubProject(r.json())

    def project_github_create(self, workspace, project, github_owner, github_repo, **kwargs):
        return SavviHubProject(self.post(f'/api/v1/workspaces/{workspace}/projects_github', {
            'name': project,
            'github_owner': github_owner,
            'github_repo': github_repo,
        }, **kwargs).json())

    def public_dataset_list(self, **kwargs):
        results = self.get_all(f'/api/v1/datasets/public', **kwargs)
        return [SavviHubDataset(x) for x in results]

    def dataset_list(self, workspace, **kwargs):
        results = self.get_all(f'/api/v1/workspaces/{workspace}/datasets', **kwargs)
        if not results:
            return []
        return [SavviHubDataset(x) for x in results]

    def dataset_read(self, workspace, dataset, **kwargs):
        r = self.get(f'/api/v1/workspaces/{workspace}/datasets/{dataset}', **kwargs)
        if r.status_code != 200:
            return None
        return SavviHubDataset(r.json())

    def dataset_snapshot_file_list(self, workspace, dataset, ref, **kwargs):
        r = self.get(f'/api/v1/workspaces/{workspace}/datasets/{dataset}/snapshots/{ref}/files', **kwargs)
        return [SavviHubFileObject(x) for x in r.json().get('results')]

    def dataset_create(self, workspace, name, is_public, description, **kwargs):
        r = self.post(f'/api/v1/workspaces/{workspace}/datasets', {
            'name': name,
            'is_public': is_public,
            'description': description,
        }, **kwargs)
        if r.status_code == 409:
            typer.echo(f"Duplicate entity \"{workspace}/{name}\". Try another dataset name!")
            return
        elif r.status_code != 200:
            typer.echo('Unexpected error.')
            return
        return SavviHubDataset(r.json())

    def dataset_gs_create(self, workspace, name, is_public, description, gs_path, **kwargs):
        r = self.post(f'/api/v1/workspaces/{workspace}/datasets_gs', {
            'name': name,
            'is_public': is_public,
            'description': description,
            'gs_path': gs_path
        }, **kwargs)
        if r.status_code == 409:
            typer.echo(f"Duplicate entity \"{workspace}/{name}\". Try another dataset name!")
            return
        return SavviHubDataset(r.json())

    def dataset_s3_create(self, workspace, name, is_public, description, s3_path, aws_role_arn, **kwargs):
        r = self.post(f'/api/v1/workspaces/{workspace}/datasets_s3', {
            'name': name,
            'is_public': is_public,
            'description': description,
            's3_path': s3_path,
            'aws_role_arn': aws_role_arn,
        }, **kwargs)
        if r.status_code == 409:
            typer.echo(f"Duplicate entity \"{workspace}/{name}\". Try another dataset name!")
            return
        return SavviHubDataset(r.json())

    def cluster_create(self, workspace, cluster_name):
        try:
            r = self.post(f'/api/v1/workspaces/{workspace}/custom_clusters', {
                'name': cluster_name,
            }, raise_error=True)
        except SavviHubError as e:
            typer.echo(e.message())
            sys.exit(1)

        return SavviHubKernelCluster(r.json())

    def cluster_add(self, workspace_name, cluster_name, agent_access_token,
                    master_endpoint, namespace, service_account_token, ssl_ca_cert_base64_encoded):
        try:
            r = self.post(f'/api/v1/workspaces/{workspace_name}/custom_clusters', {
                'name': cluster_name,
                'agent_access_token': agent_access_token,
                'kubernetes_master_endpoint': master_endpoint,
                'kubernetes_namespace': namespace,
                'kubernetes_service_account_token': service_account_token,
                'kubernetes_ssl_ca_cert': ssl_ca_cert_base64_encoded,
            }, raise_error=True)
        except SavviHubError as e:
            typer.echo(e.message())
            sys.exit(1)

        return SavviHubKernelCluster(r.json())

    def cluster_list(self, workspace_name):
        try:
            r = self.get(f'/api/v1/workspaces/{workspace_name}/clusters', raise_error=True)
        except SavviHubError as e:
            typer.echo(e.message())
            sys.exit(1)

        return [SavviHubKernelCluster(x) for x in r.json()['clusters']]

    def cluster_delete(self, workspace_name, cluster_name):
        try:
            self.delete(f'/api/v1/workspaces/{workspace_name}/custom_clusters/{cluster_name}', raise_error=True)
        except SavviHubError as e:
            typer.echo(e.message())
            sys.exit(1)

    def cluster_rename(self, workspace_name, cluster_name, new_name):
        try:
            self.patch(f'/api/v1/workspaces/{workspace_name}/custom_clusters/{cluster_name}', {
                'name': new_name,
            }, raise_error=True)
        except SavviHubError as e:
            typer.echo(e.message())
            sys.exit(1)

    @staticmethod
    def get_full_info(workspace, project, experiment, form):
        return \
            f'\n\tFull {form} info at:\n' \
            f'\t\t{WEB_HOST}/{workspace}/{project}/experiments/{experiment}/{form}\n'

    @staticmethod
    def get_full_info_project(workspace, project):
        return \
            f'\n\tFull project info at:\n' \
            f'\t\t{WEB_HOST}/{workspace}/{project}\n'

    @staticmethod
    def get_full_info_dataset(workspace, dataset):
        return \
            f'\n\tFull dataset info at:\n' \
            f'\t\t{WEB_HOST}/{workspace}/datasets/{dataset}\n'

    @staticmethod
    def get_full_info_experiment(workspace, project, experiment):
        return \
            f'\n\tFull experiment info at:\n' \
            f'\t\t{WEB_HOST}/{workspace}/{project}/experiments/{experiment}\n'

    @staticmethod
    def get_download_url(volume_id, path):
        return f'{WEB_HOST}/file-download/{volume_id}/{path}'



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


import inspect
import pprint
import re  # noqa: F401
import six

from savvihub_client.configuration import Configuration


class ModelResourceSpecFields(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'cpu_limit': 'float',
        'cpu_type': 'str',
        'gpu_limit': 'int',
        'gpu_type': 'str',
        'memory_limit': 'str',
        'processor_type': 'str'
    }

    attribute_map = {
        'cpu_limit': 'cpu_limit',
        'cpu_type': 'cpu_type',
        'gpu_limit': 'gpu_limit',
        'gpu_type': 'gpu_type',
        'memory_limit': 'memory_limit',
        'processor_type': 'processor_type'
    }

    def __init__(self, cpu_limit=None, cpu_type=None, gpu_limit=None, gpu_type=None, memory_limit=None, processor_type=None, local_vars_configuration=None):  # noqa: E501
        """ModelResourceSpecFields - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._cpu_limit = None
        self._cpu_type = None
        self._gpu_limit = None
        self._gpu_type = None
        self._memory_limit = None
        self._processor_type = None
        self.discriminator = None

        self.cpu_limit = cpu_limit
        self.cpu_type = cpu_type
        if gpu_limit is not None:
            self.gpu_limit = gpu_limit
        self.gpu_type = gpu_type
        self.memory_limit = memory_limit
        self.processor_type = processor_type

    @property
    def cpu_limit(self):
        """Gets the cpu_limit of this ModelResourceSpecFields.  # noqa: E501


        :return: The cpu_limit of this ModelResourceSpecFields.  # noqa: E501
        :rtype: float
        """
        return self._cpu_limit

    @cpu_limit.setter
    def cpu_limit(self, cpu_limit):
        """Sets the cpu_limit of this ModelResourceSpecFields.


        :param cpu_limit: The cpu_limit of this ModelResourceSpecFields.  # noqa: E501
        :type cpu_limit: float
        """
        if self.local_vars_configuration.client_side_validation and cpu_limit is None:  # noqa: E501
            raise ValueError("Invalid value for `cpu_limit`, must not be `None`")  # noqa: E501

        self._cpu_limit = cpu_limit

    @property
    def cpu_type(self):
        """Gets the cpu_type of this ModelResourceSpecFields.  # noqa: E501


        :return: The cpu_type of this ModelResourceSpecFields.  # noqa: E501
        :rtype: str
        """
        return self._cpu_type

    @cpu_type.setter
    def cpu_type(self, cpu_type):
        """Sets the cpu_type of this ModelResourceSpecFields.


        :param cpu_type: The cpu_type of this ModelResourceSpecFields.  # noqa: E501
        :type cpu_type: str
        """
        if self.local_vars_configuration.client_side_validation and cpu_type is None:  # noqa: E501
            raise ValueError("Invalid value for `cpu_type`, must not be `None`")  # noqa: E501

        self._cpu_type = cpu_type

    @property
    def gpu_limit(self):
        """Gets the gpu_limit of this ModelResourceSpecFields.  # noqa: E501


        :return: The gpu_limit of this ModelResourceSpecFields.  # noqa: E501
        :rtype: int
        """
        return self._gpu_limit

    @gpu_limit.setter
    def gpu_limit(self, gpu_limit):
        """Sets the gpu_limit of this ModelResourceSpecFields.


        :param gpu_limit: The gpu_limit of this ModelResourceSpecFields.  # noqa: E501
        :type gpu_limit: int
        """

        self._gpu_limit = gpu_limit

    @property
    def gpu_type(self):
        """Gets the gpu_type of this ModelResourceSpecFields.  # noqa: E501


        :return: The gpu_type of this ModelResourceSpecFields.  # noqa: E501
        :rtype: str
        """
        return self._gpu_type

    @gpu_type.setter
    def gpu_type(self, gpu_type):
        """Sets the gpu_type of this ModelResourceSpecFields.


        :param gpu_type: The gpu_type of this ModelResourceSpecFields.  # noqa: E501
        :type gpu_type: str
        """
        if self.local_vars_configuration.client_side_validation and gpu_type is None:  # noqa: E501
            raise ValueError("Invalid value for `gpu_type`, must not be `None`")  # noqa: E501

        self._gpu_type = gpu_type

    @property
    def memory_limit(self):
        """Gets the memory_limit of this ModelResourceSpecFields.  # noqa: E501


        :return: The memory_limit of this ModelResourceSpecFields.  # noqa: E501
        :rtype: str
        """
        return self._memory_limit

    @memory_limit.setter
    def memory_limit(self, memory_limit):
        """Sets the memory_limit of this ModelResourceSpecFields.


        :param memory_limit: The memory_limit of this ModelResourceSpecFields.  # noqa: E501
        :type memory_limit: str
        """
        if self.local_vars_configuration.client_side_validation and memory_limit is None:  # noqa: E501
            raise ValueError("Invalid value for `memory_limit`, must not be `None`")  # noqa: E501

        self._memory_limit = memory_limit

    @property
    def processor_type(self):
        """Gets the processor_type of this ModelResourceSpecFields.  # noqa: E501


        :return: The processor_type of this ModelResourceSpecFields.  # noqa: E501
        :rtype: str
        """
        return self._processor_type

    @processor_type.setter
    def processor_type(self, processor_type):
        """Sets the processor_type of this ModelResourceSpecFields.


        :param processor_type: The processor_type of this ModelResourceSpecFields.  # noqa: E501
        :type processor_type: str
        """
        if self.local_vars_configuration.client_side_validation and processor_type is None:  # noqa: E501
            raise ValueError("Invalid value for `processor_type`, must not be `None`")  # noqa: E501

        self._processor_type = processor_type

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = inspect.getargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ModelResourceSpecFields):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ModelResourceSpecFields):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


import inspect
import pprint
import re  # noqa: F401
import six

from savvihub_client.configuration import Configuration


class ResponseExperiment(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'created_by': 'ModelUser',
        'created_by_id': 'int',
        'created_dt': 'datetime',
        'env_vars': 'list[ModelEnvVar]',
        'git_diff_file': 'ResponseFileMetadata',
        'git_ref': 'str',
        'histories': 'list[ResponseExperimentHistory]',
        'id': 'int',
        'kernel_cluster': 'ResponseExperimentKernelCluster',
        'kernel_cluster_node': 'ResponseKernelClusterNode',
        'kernel_image': 'ModelKernelImage',
        'kernel_image_id': 'int',
        'kernel_resource_spec': 'ResponseKernelResourceSpec',
        'kernel_resource_spec_id': 'int',
        'message': 'str',
        'metrics_summary': 'ModelExperimentMetricsSummary',
        'name': 'str',
        'number': 'int',
        'project_id': 'int',
        'reproduce_command': 'str',
        'slug': 'str',
        'source_code_link': 'list[ResponseProjectLink]',
        'start_command': 'str',
        'status': 'str',
        'tensorboard': 'ResponseService',
        'updated_dt': 'datetime',
        'volumes': 'list[ResponseVolumeMountInfo]',
        'workspace_id': 'int'
    }

    attribute_map = {
        'created_by': 'created_by',
        'created_by_id': 'created_by_id',
        'created_dt': 'created_dt',
        'env_vars': 'env_vars',
        'git_diff_file': 'git_diff_file',
        'git_ref': 'git_ref',
        'histories': 'histories',
        'id': 'id',
        'kernel_cluster': 'kernel_cluster',
        'kernel_cluster_node': 'kernel_cluster_node',
        'kernel_image': 'kernel_image',
        'kernel_image_id': 'kernel_image_id',
        'kernel_resource_spec': 'kernel_resource_spec',
        'kernel_resource_spec_id': 'kernel_resource_spec_id',
        'message': 'message',
        'metrics_summary': 'metrics_summary',
        'name': 'name',
        'number': 'number',
        'project_id': 'project_id',
        'reproduce_command': 'reproduce_command',
        'slug': 'slug',
        'source_code_link': 'source_code_link',
        'start_command': 'start_command',
        'status': 'status',
        'tensorboard': 'tensorboard',
        'updated_dt': 'updated_dt',
        'volumes': 'volumes',
        'workspace_id': 'workspace_id'
    }

    def __init__(self, created_by=None, created_by_id=None, created_dt=None, env_vars=None, git_diff_file=None, git_ref=None, histories=None, id=None, kernel_cluster=None, kernel_cluster_node=None, kernel_image=None, kernel_image_id=None, kernel_resource_spec=None, kernel_resource_spec_id=None, message=None, metrics_summary=None, name=None, number=None, project_id=None, reproduce_command=None, slug=None, source_code_link=None, start_command=None, status=None, tensorboard=None, updated_dt=None, volumes=None, workspace_id=None, local_vars_configuration=None):  # noqa: E501
        """ResponseExperiment - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._created_by = None
        self._created_by_id = None
        self._created_dt = None
        self._env_vars = None
        self._git_diff_file = None
        self._git_ref = None
        self._histories = None
        self._id = None
        self._kernel_cluster = None
        self._kernel_cluster_node = None
        self._kernel_image = None
        self._kernel_image_id = None
        self._kernel_resource_spec = None
        self._kernel_resource_spec_id = None
        self._message = None
        self._metrics_summary = None
        self._name = None
        self._number = None
        self._project_id = None
        self._reproduce_command = None
        self._slug = None
        self._source_code_link = None
        self._start_command = None
        self._status = None
        self._tensorboard = None
        self._updated_dt = None
        self._volumes = None
        self._workspace_id = None
        self.discriminator = None

        self.created_by = created_by
        self.created_by_id = created_by_id
        self.created_dt = created_dt
        self.env_vars = env_vars
        if git_diff_file is not None:
            self.git_diff_file = git_diff_file
        if git_ref is not None:
            self.git_ref = git_ref
        self.histories = histories
        self.id = id
        if kernel_cluster is not None:
            self.kernel_cluster = kernel_cluster
        if kernel_cluster_node is not None:
            self.kernel_cluster_node = kernel_cluster_node
        self.kernel_image = kernel_image
        self.kernel_image_id = kernel_image_id
        self.kernel_resource_spec = kernel_resource_spec
        self.kernel_resource_spec_id = kernel_resource_spec_id
        self.message = message
        self.metrics_summary = metrics_summary
        self.name = name
        self.number = number
        self.project_id = project_id
        if reproduce_command is not None:
            self.reproduce_command = reproduce_command
        self.slug = slug
        self.source_code_link = source_code_link
        self.start_command = start_command
        self.status = status
        self.tensorboard = tensorboard
        self.updated_dt = updated_dt
        self.volumes = volumes
        self.workspace_id = workspace_id

    @property
    def created_by(self):
        """Gets the created_by of this ResponseExperiment.  # noqa: E501


        :return: The created_by of this ResponseExperiment.  # noqa: E501
        :rtype: ModelUser
        """
        return self._created_by

    @created_by.setter
    def created_by(self, created_by):
        """Sets the created_by of this ResponseExperiment.


        :param created_by: The created_by of this ResponseExperiment.  # noqa: E501
        :type created_by: ModelUser
        """
        if self.local_vars_configuration.client_side_validation and created_by is None:  # noqa: E501
            raise ValueError("Invalid value for `created_by`, must not be `None`")  # noqa: E501

        self._created_by = created_by

    @property
    def created_by_id(self):
        """Gets the created_by_id of this ResponseExperiment.  # noqa: E501


        :return: The created_by_id of this ResponseExperiment.  # noqa: E501
        :rtype: int
        """
        return self._created_by_id

    @created_by_id.setter
    def created_by_id(self, created_by_id):
        """Sets the created_by_id of this ResponseExperiment.


        :param created_by_id: The created_by_id of this ResponseExperiment.  # noqa: E501
        :type created_by_id: int
        """
        if self.local_vars_configuration.client_side_validation and created_by_id is None:  # noqa: E501
            raise ValueError("Invalid value for `created_by_id`, must not be `None`")  # noqa: E501

        self._created_by_id = created_by_id

    @property
    def created_dt(self):
        """Gets the created_dt of this ResponseExperiment.  # noqa: E501


        :return: The created_dt of this ResponseExperiment.  # noqa: E501
        :rtype: datetime
        """
        return self._created_dt

    @created_dt.setter
    def created_dt(self, created_dt):
        """Sets the created_dt of this ResponseExperiment.


        :param created_dt: The created_dt of this ResponseExperiment.  # noqa: E501
        :type created_dt: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_dt is None:  # noqa: E501
            raise ValueError("Invalid value for `created_dt`, must not be `None`")  # noqa: E501

        self._created_dt = created_dt

    @property
    def env_vars(self):
        """Gets the env_vars of this ResponseExperiment.  # noqa: E501


        :return: The env_vars of this ResponseExperiment.  # noqa: E501
        :rtype: list[ModelEnvVar]
        """
        return self._env_vars

    @env_vars.setter
    def env_vars(self, env_vars):
        """Sets the env_vars of this ResponseExperiment.


        :param env_vars: The env_vars of this ResponseExperiment.  # noqa: E501
        :type env_vars: list[ModelEnvVar]
        """
        if self.local_vars_configuration.client_side_validation and env_vars is None:  # noqa: E501
            raise ValueError("Invalid value for `env_vars`, must not be `None`")  # noqa: E501

        self._env_vars = env_vars

    @property
    def git_diff_file(self):
        """Gets the git_diff_file of this ResponseExperiment.  # noqa: E501


        :return: The git_diff_file of this ResponseExperiment.  # noqa: E501
        :rtype: ResponseFileMetadata
        """
        return self._git_diff_file

    @git_diff_file.setter
    def git_diff_file(self, git_diff_file):
        """Sets the git_diff_file of this ResponseExperiment.


        :param git_diff_file: The git_diff_file of this ResponseExperiment.  # noqa: E501
        :type git_diff_file: ResponseFileMetadata
        """

        self._git_diff_file = git_diff_file

    @property
    def git_ref(self):
        """Gets the git_ref of this ResponseExperiment.  # noqa: E501


        :return: The git_ref of this ResponseExperiment.  # noqa: E501
        :rtype: str
        """
        return self._git_ref

    @git_ref.setter
    def git_ref(self, git_ref):
        """Sets the git_ref of this ResponseExperiment.


        :param git_ref: The git_ref of this ResponseExperiment.  # noqa: E501
        :type git_ref: str
        """

        self._git_ref = git_ref

    @property
    def histories(self):
        """Gets the histories of this ResponseExperiment.  # noqa: E501


        :return: The histories of this ResponseExperiment.  # noqa: E501
        :rtype: list[ResponseExperimentHistory]
        """
        return self._histories

    @histories.setter
    def histories(self, histories):
        """Sets the histories of this ResponseExperiment.


        :param histories: The histories of this ResponseExperiment.  # noqa: E501
        :type histories: list[ResponseExperimentHistory]
        """
        if self.local_vars_configuration.client_side_validation and histories is None:  # noqa: E501
            raise ValueError("Invalid value for `histories`, must not be `None`")  # noqa: E501

        self._histories = histories

    @property
    def id(self):
        """Gets the id of this ResponseExperiment.  # noqa: E501


        :return: The id of this ResponseExperiment.  # noqa: E501
        :rtype: int
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this ResponseExperiment.


        :param id: The id of this ResponseExperiment.  # noqa: E501
        :type id: int
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def kernel_cluster(self):
        """Gets the kernel_cluster of this ResponseExperiment.  # noqa: E501


        :return: The kernel_cluster of this ResponseExperiment.  # noqa: E501
        :rtype: ResponseExperimentKernelCluster
        """
        return self._kernel_cluster

    @kernel_cluster.setter
    def kernel_cluster(self, kernel_cluster):
        """Sets the kernel_cluster of this ResponseExperiment.


        :param kernel_cluster: The kernel_cluster of this ResponseExperiment.  # noqa: E501
        :type kernel_cluster: ResponseExperimentKernelCluster
        """

        self._kernel_cluster = kernel_cluster

    @property
    def kernel_cluster_node(self):
        """Gets the kernel_cluster_node of this ResponseExperiment.  # noqa: E501


        :return: The kernel_cluster_node of this ResponseExperiment.  # noqa: E501
        :rtype: ResponseKernelClusterNode
        """
        return self._kernel_cluster_node

    @kernel_cluster_node.setter
    def kernel_cluster_node(self, kernel_cluster_node):
        """Sets the kernel_cluster_node of this ResponseExperiment.


        :param kernel_cluster_node: The kernel_cluster_node of this ResponseExperiment.  # noqa: E501
        :type kernel_cluster_node: ResponseKernelClusterNode
        """

        self._kernel_cluster_node = kernel_cluster_node

    @property
    def kernel_image(self):
        """Gets the kernel_image of this ResponseExperiment.  # noqa: E501


        :return: The kernel_image of this ResponseExperiment.  # noqa: E501
        :rtype: ModelKernelImage
        """
        return self._kernel_image

    @kernel_image.setter
    def kernel_image(self, kernel_image):
        """Sets the kernel_image of this ResponseExperiment.


        :param kernel_image: The kernel_image of this ResponseExperiment.  # noqa: E501
        :type kernel_image: ModelKernelImage
        """
        if self.local_vars_configuration.client_side_validation and kernel_image is None:  # noqa: E501
            raise ValueError("Invalid value for `kernel_image`, must not be `None`")  # noqa: E501

        self._kernel_image = kernel_image

    @property
    def kernel_image_id(self):
        """Gets the kernel_image_id of this ResponseExperiment.  # noqa: E501


        :return: The kernel_image_id of this ResponseExperiment.  # noqa: E501
        :rtype: int
        """
        return self._kernel_image_id

    @kernel_image_id.setter
    def kernel_image_id(self, kernel_image_id):
        """Sets the kernel_image_id of this ResponseExperiment.


        :param kernel_image_id: The kernel_image_id of this ResponseExperiment.  # noqa: E501
        :type kernel_image_id: int
        """
        if self.local_vars_configuration.client_side_validation and kernel_image_id is None:  # noqa: E501
            raise ValueError("Invalid value for `kernel_image_id`, must not be `None`")  # noqa: E501

        self._kernel_image_id = kernel_image_id

    @property
    def kernel_resource_spec(self):
        """Gets the kernel_resource_spec of this ResponseExperiment.  # noqa: E501


        :return: The kernel_resource_spec of this ResponseExperiment.  # noqa: E501
        :rtype: ResponseKernelResourceSpec
        """
        return self._kernel_resource_spec

    @kernel_resource_spec.setter
    def kernel_resource_spec(self, kernel_resource_spec):
        """Sets the kernel_resource_spec of this ResponseExperiment.


        :param kernel_resource_spec: The kernel_resource_spec of this ResponseExperiment.  # noqa: E501
        :type kernel_resource_spec: ResponseKernelResourceSpec
        """
        if self.local_vars_configuration.client_side_validation and kernel_resource_spec is None:  # noqa: E501
            raise ValueError("Invalid value for `kernel_resource_spec`, must not be `None`")  # noqa: E501

        self._kernel_resource_spec = kernel_resource_spec

    @property
    def kernel_resource_spec_id(self):
        """Gets the kernel_resource_spec_id of this ResponseExperiment.  # noqa: E501


        :return: The kernel_resource_spec_id of this ResponseExperiment.  # noqa: E501
        :rtype: int
        """
        return self._kernel_resource_spec_id

    @kernel_resource_spec_id.setter
    def kernel_resource_spec_id(self, kernel_resource_spec_id):
        """Sets the kernel_resource_spec_id of this ResponseExperiment.


        :param kernel_resource_spec_id: The kernel_resource_spec_id of this ResponseExperiment.  # noqa: E501
        :type kernel_resource_spec_id: int
        """

        self._kernel_resource_spec_id = kernel_resource_spec_id

    @property
    def message(self):
        """Gets the message of this ResponseExperiment.  # noqa: E501


        :return: The message of this ResponseExperiment.  # noqa: E501
        :rtype: str
        """
        return self._message

    @message.setter
    def message(self, message):
        """Sets the message of this ResponseExperiment.


        :param message: The message of this ResponseExperiment.  # noqa: E501
        :type message: str
        """
        if self.local_vars_configuration.client_side_validation and message is None:  # noqa: E501
            raise ValueError("Invalid value for `message`, must not be `None`")  # noqa: E501

        self._message = message

    @property
    def metrics_summary(self):
        """Gets the metrics_summary of this ResponseExperiment.  # noqa: E501


        :return: The metrics_summary of this ResponseExperiment.  # noqa: E501
        :rtype: ModelExperimentMetricsSummary
        """
        return self._metrics_summary

    @metrics_summary.setter
    def metrics_summary(self, metrics_summary):
        """Sets the metrics_summary of this ResponseExperiment.


        :param metrics_summary: The metrics_summary of this ResponseExperiment.  # noqa: E501
        :type metrics_summary: ModelExperimentMetricsSummary
        """
        if self.local_vars_configuration.client_side_validation and metrics_summary is None:  # noqa: E501
            raise ValueError("Invalid value for `metrics_summary`, must not be `None`")  # noqa: E501

        self._metrics_summary = metrics_summary

    @property
    def name(self):
        """Gets the name of this ResponseExperiment.  # noqa: E501


        :return: The name of this ResponseExperiment.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this ResponseExperiment.


        :param name: The name of this ResponseExperiment.  # noqa: E501
        :type name: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def number(self):
        """Gets the number of this ResponseExperiment.  # noqa: E501


        :return: The number of this ResponseExperiment.  # noqa: E501
        :rtype: int
        """
        return self._number

    @number.setter
    def number(self, number):
        """Sets the number of this ResponseExperiment.


        :param number: The number of this ResponseExperiment.  # noqa: E501
        :type number: int
        """
        if self.local_vars_configuration.client_side_validation and number is None:  # noqa: E501
            raise ValueError("Invalid value for `number`, must not be `None`")  # noqa: E501

        self._number = number

    @property
    def project_id(self):
        """Gets the project_id of this ResponseExperiment.  # noqa: E501


        :return: The project_id of this ResponseExperiment.  # noqa: E501
        :rtype: int
        """
        return self._project_id

    @project_id.setter
    def project_id(self, project_id):
        """Sets the project_id of this ResponseExperiment.


        :param project_id: The project_id of this ResponseExperiment.  # noqa: E501
        :type project_id: int
        """
        if self.local_vars_configuration.client_side_validation and project_id is None:  # noqa: E501
            raise ValueError("Invalid value for `project_id`, must not be `None`")  # noqa: E501

        self._project_id = project_id

    @property
    def reproduce_command(self):
        """Gets the reproduce_command of this ResponseExperiment.  # noqa: E501


        :return: The reproduce_command of this ResponseExperiment.  # noqa: E501
        :rtype: str
        """
        return self._reproduce_command

    @reproduce_command.setter
    def reproduce_command(self, reproduce_command):
        """Sets the reproduce_command of this ResponseExperiment.


        :param reproduce_command: The reproduce_command of this ResponseExperiment.  # noqa: E501
        :type reproduce_command: str
        """

        self._reproduce_command = reproduce_command

    @property
    def slug(self):
        """Gets the slug of this ResponseExperiment.  # noqa: E501


        :return: The slug of this ResponseExperiment.  # noqa: E501
        :rtype: str
        """
        return self._slug

    @slug.setter
    def slug(self, slug):
        """Sets the slug of this ResponseExperiment.


        :param slug: The slug of this ResponseExperiment.  # noqa: E501
        :type slug: str
        """
        if self.local_vars_configuration.client_side_validation and slug is None:  # noqa: E501
            raise ValueError("Invalid value for `slug`, must not be `None`")  # noqa: E501

        self._slug = slug

    @property
    def source_code_link(self):
        """Gets the source_code_link of this ResponseExperiment.  # noqa: E501


        :return: The source_code_link of this ResponseExperiment.  # noqa: E501
        :rtype: list[ResponseProjectLink]
        """
        return self._source_code_link

    @source_code_link.setter
    def source_code_link(self, source_code_link):
        """Sets the source_code_link of this ResponseExperiment.


        :param source_code_link: The source_code_link of this ResponseExperiment.  # noqa: E501
        :type source_code_link: list[ResponseProjectLink]
        """
        if self.local_vars_configuration.client_side_validation and source_code_link is None:  # noqa: E501
            raise ValueError("Invalid value for `source_code_link`, must not be `None`")  # noqa: E501

        self._source_code_link = source_code_link

    @property
    def start_command(self):
        """Gets the start_command of this ResponseExperiment.  # noqa: E501


        :return: The start_command of this ResponseExperiment.  # noqa: E501
        :rtype: str
        """
        return self._start_command

    @start_command.setter
    def start_command(self, start_command):
        """Sets the start_command of this ResponseExperiment.


        :param start_command: The start_command of this ResponseExperiment.  # noqa: E501
        :type start_command: str
        """
        if self.local_vars_configuration.client_side_validation and start_command is None:  # noqa: E501
            raise ValueError("Invalid value for `start_command`, must not be `None`")  # noqa: E501

        self._start_command = start_command

    @property
    def status(self):
        """Gets the status of this ResponseExperiment.  # noqa: E501


        :return: The status of this ResponseExperiment.  # noqa: E501
        :rtype: str
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this ResponseExperiment.


        :param status: The status of this ResponseExperiment.  # noqa: E501
        :type status: str
        """
        if self.local_vars_configuration.client_side_validation and status is None:  # noqa: E501
            raise ValueError("Invalid value for `status`, must not be `None`")  # noqa: E501

        self._status = status

    @property
    def tensorboard(self):
        """Gets the tensorboard of this ResponseExperiment.  # noqa: E501


        :return: The tensorboard of this ResponseExperiment.  # noqa: E501
        :rtype: ResponseService
        """
        return self._tensorboard

    @tensorboard.setter
    def tensorboard(self, tensorboard):
        """Sets the tensorboard of this ResponseExperiment.


        :param tensorboard: The tensorboard of this ResponseExperiment.  # noqa: E501
        :type tensorboard: ResponseService
        """
        if self.local_vars_configuration.client_side_validation and tensorboard is None:  # noqa: E501
            raise ValueError("Invalid value for `tensorboard`, must not be `None`")  # noqa: E501

        self._tensorboard = tensorboard

    @property
    def updated_dt(self):
        """Gets the updated_dt of this ResponseExperiment.  # noqa: E501


        :return: The updated_dt of this ResponseExperiment.  # noqa: E501
        :rtype: datetime
        """
        return self._updated_dt

    @updated_dt.setter
    def updated_dt(self, updated_dt):
        """Sets the updated_dt of this ResponseExperiment.


        :param updated_dt: The updated_dt of this ResponseExperiment.  # noqa: E501
        :type updated_dt: datetime
        """
        if self.local_vars_configuration.client_side_validation and updated_dt is None:  # noqa: E501
            raise ValueError("Invalid value for `updated_dt`, must not be `None`")  # noqa: E501

        self._updated_dt = updated_dt

    @property
    def volumes(self):
        """Gets the volumes of this ResponseExperiment.  # noqa: E501


        :return: The volumes of this ResponseExperiment.  # noqa: E501
        :rtype: list[ResponseVolumeMountInfo]
        """
        return self._volumes

    @volumes.setter
    def volumes(self, volumes):
        """Sets the volumes of this ResponseExperiment.


        :param volumes: The volumes of this ResponseExperiment.  # noqa: E501
        :type volumes: list[ResponseVolumeMountInfo]
        """
        if self.local_vars_configuration.client_side_validation and volumes is None:  # noqa: E501
            raise ValueError("Invalid value for `volumes`, must not be `None`")  # noqa: E501

        self._volumes = volumes

    @property
    def workspace_id(self):
        """Gets the workspace_id of this ResponseExperiment.  # noqa: E501


        :return: The workspace_id of this ResponseExperiment.  # noqa: E501
        :rtype: int
        """
        return self._workspace_id

    @workspace_id.setter
    def workspace_id(self, workspace_id):
        """Sets the workspace_id of this ResponseExperiment.


        :param workspace_id: The workspace_id of this ResponseExperiment.  # noqa: E501
        :type workspace_id: int
        """
        if self.local_vars_configuration.client_side_validation and workspace_id is None:  # noqa: E501
            raise ValueError("Invalid value for `workspace_id`, must not be `None`")  # noqa: E501

        self._workspace_id = workspace_id

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = inspect.getargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ResponseExperiment):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ResponseExperiment):
            return True

        return self.to_dict() != other.to_dict()



import base64
import subprocess
import sys
from functools import wraps
from typing import Optional, Tuple

import typer
import yaml


def kubectl(f):
    @wraps(f)
    def wrapped(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except FileNotFoundError as e:
            if 'kubectl' in str(e):
                typer.echo('\'kubectl\' must be available. See https://kubernetes.io/docs/tasks/tools/install-kubectl.')
            else:
                typer.echo(str(e))
            sys.exit(1)
        except subprocess.CalledProcessError:
            sys.exit(1)

    return wrapped


@kubectl
def check_context(context: str) -> bool:
    return context == subprocess.check_output(['kubectl', 'config', 'current-context']).decode().strip()


@kubectl
def get_current_context() -> str:
    return subprocess.check_output(['kubectl', 'config', 'current-context']).decode().strip()


@kubectl
def get_cluster_info(context: str) -> Optional[Tuple[str, str]]:
    configs = yaml.load(subprocess.check_output(['kubectl', 'config', 'view', '--raw']).decode().strip(), yaml.FullLoader)
    for cluster_info in configs['clusters']:
        if cluster_info['name'] == context:
            cluster_info = cluster_info['cluster']
            ssl_ca_cert_base64_encoded = ''
            if 'certificate-authority' in cluster_info:
                with open(cluster_info['certificate-authority'], 'r') as f:
                    ssl_ca_cert_base64_encoded = base64.b64encode(f.read().encode()).decode()
            elif 'certificate-authority-data' in cluster_info:
                ssl_ca_cert_base64_encoded = cluster_info['certificate-authority-data']

            return cluster_info['server'], ssl_ca_cert_base64_encoded

    typer.echo(f'No such context: {context}')
    sys.exit(1)


@kubectl
def get_service_account_token(namespace: str) -> Optional[str]:
    try:
        service_account = yaml.load(subprocess.check_output(['kubectl', 'get', 'serviceaccounts/savvihub', '-n', namespace, '-o', 'yaml']), yaml.FullLoader)
    except subprocess.CalledProcessError:
        sys.exit(1)

    secret_name = service_account.get('secrets', [{}])[0].get('name')
    if secret_name is None:
        typer.echo('The service account does not have a required secret token.')
        sys.exit(1)

    try:
        secret = yaml.load(subprocess.check_output(['kubectl', 'get', 'secrets', secret_name, '-n', namespace, '-o', 'yaml']), yaml.FullLoader)
    except subprocess.CalledProcessError:
        sys.exit(1)

    return base64.b64decode(secret['data']['token']).decode()


@kubectl
def get_agent_access_token(namespace: str) -> Optional[str]:
    secret = yaml.load(
        subprocess.check_output(['kubectl', 'get', 'secrets', 'savvihub-agent', '-n', namespace, '-o', 'yaml']),
        yaml.FullLoader,
    )
    return base64.b64decode(secret['data']['access-token']).decode()



import os
import pathlib
import re
import sys
import tempfile
from typing import List

import inquirer
import typer

from terminaltables import AsciiTable

from savvihub.api.file_object import DownloadableFileObject
from savvihub.api.uploader import Uploader
from savvihub.common.constants import INQUIRER_NAME_COMMAND, WEB_HOST, API_HOST, INQUIRER_NAME_CPU_LIMIT, \
    INQUIRER_NAME_MEMORY_LIMIT, INQUIRER_NAME_GPU_LIMIT
from savvihub.common.context import Context
from savvihub.common.formatter import TreeFormatter
from savvihub.common.utils import short_string, parse_time_to_ago, parse_str_time_to_datetime, \
    parse_timestamp_or_none, remove_file
from savvihub.savvihub_cli.errors import get_error_message
from savvihub.savvihub_cli.volume import volume_file_list, volume_file_copy

experiment_output_app = typer.Typer()
experiment_app = typer.Typer()
experiment_app.add_typer(experiment_output_app, name='output')


@experiment_app.callback()
def main():
    """
    Run the machine learning experiment
    """
    return


@experiment_app.command()
def list():
    """
    Display a list of experiments
    """

    context = Context(user_required=True, project_required=True)
    client = context.authorized_client

    experiments = client.experiment_list(context.project.workspace.name, context.project.name, raise_error=True)
    if not experiments:
        typer.echo(f'There is no experiment in {context.project.name} project.')
        return

    table = AsciiTable([
        ['NUMBER', 'NAME', 'STATUS', 'CREATED', 'IMAGE', 'RESOURCE', 'START COMMAND'],
        *[[e.number, e.name, e.status, parse_time_to_ago(e.created_dt),  short_string(e.kernel_image.name, 25),
           e.kernel_resource_spec.name, f'"{short_string(e.start_command, 25)}"']
          for e in experiments],
    ])
    table.inner_column_border = False
    table.inner_heading_row_border = False
    table.inner_footing_row_border = False
    table.outer_border = False

    typer.echo(table.table)


@experiment_app.command()
def describe(
    experiment_number_or_name: str = typer.Argument(..., help="The unique experiment number or name"),
):
    """
    Describe the experiment in details
    """
    context = Context(user_required=True, project_required=True)
    client = context.authorized_client

    experiment = client.experiment_read(
        context.project.workspace.name, context.project.name, experiment_number_or_name,
        raise_error=True,
    )

    root = TreeFormatter()
    root.add_child(f'Number: {experiment.number}',
                   f'Name: {experiment.name}',
                   f'Created: {parse_str_time_to_datetime(experiment.created_dt)}',
                   f'Updated: {parse_str_time_to_datetime(experiment.updated_dt)}',
                   f'Git Commit Message: ({experiment.git_ref}) {experiment.message}')

    if experiment.git_diff_file:
        git_diff_file_desc = TreeFormatter('Git Diff File:')
        git_diff_file_desc.add_child(f'URL: {experiment.git_diff_file.download_url["url"]}')
        root.add_child(git_diff_file_desc)

    root.add_child(f'Status: {experiment.status}',
                   f'Tensorboard: {experiment.tensorboard or "N/A"}',
                   f'Tensorboard log_dir: {experiment.tensorboard_log_dir or "N/A"}')

    # Kernel Image Description
    kernel_image_desc = TreeFormatter('Kernel Image:')
    kernel_image_desc.add_child(f'Name: {experiment.kernel_image.name}',
                                f'URL: {experiment.kernel_image.image_url}',
                                f'Language: {experiment.kernel_image.language}')
    root.add_child(kernel_image_desc)

    # Resource Spec Description
    resource_spec_desc = TreeFormatter('Resource Spec:')
    resource_spec_desc.add_child(f'Name: {experiment.kernel_resource_spec.name}',
                                 f'CPU Type: {experiment.kernel_resource_spec.name}',
                                 f'CPU Limit: {experiment.kernel_resource_spec.cpu_limit}',
                                 f'CPU Guarantee: {experiment.kernel_resource_spec.cpu_guarantee}',
                                 f'Memory Limit: {experiment.kernel_resource_spec.memory_limit}',
                                 f'Memory Guarantee: {experiment.kernel_resource_spec.mem_guarantee}',
                                 f'GPU Type: {experiment.kernel_resource_spec.gpu_type}',
                                 f'GPU Limit: {experiment.kernel_resource_spec.gpu_limit}',
                                 f'GPU Guarantee: {experiment.kernel_resource_spec.gpu_guarantee}')
    root.add_child(resource_spec_desc)

    # Datasets Description
    datasets_desc = TreeFormatter('Datasets:')
    if not experiment.datasets:
        datasets_desc.add_child('Dataset Not Found')
    else:
        for dataset in experiment.datasets:
            dataset_desc = TreeFormatter(dataset['dataset_name'])
            dataset_desc.add_child(f'Mount Path: {dataset["mount_path"]}')
            datasets_desc.add_child(dataset_desc)
    root.add_child(datasets_desc)

    # Histories Description
    histories_desc = TreeFormatter('Histories:')
    if not experiment.histories:
        histories_desc.add_child('History Not Found')
    else:
        for history in experiment.histories:
            history_desc = TreeFormatter(history['status'])
            history_desc.add_child(f'Started: {parse_timestamp_or_none(history["started_timestamp"])}',
                                   f'Ended: {parse_timestamp_or_none(history["ended_timestamp"])}')
            histories_desc.add_child(history_desc)
    root.add_child(histories_desc)

    # Metrics Description
    metrics_desc = TreeFormatter('Metrics:')
    if not experiment.metrics:
        metrics_desc.add_child('Metrics Not Found')
    else:
        for history in experiment.metrics[-10:]:
            metrics_desc.add_child(history)

        full_metrics_info_desc = TreeFormatter('Full metrics info at:')
        full_metrics_info_desc.add_child(f'{WEB_HOST}/{context.project.workspace.name}/{context.project.name}/experiments/{experiment.number}/metrics')
        metrics_desc.add_child(full_metrics_info_desc)
    root.add_child(metrics_desc)

    # System Metrics Description
    system_metrics_desc = TreeFormatter('System Metrics:')
    full_system_metrics_info_desc = TreeFormatter('Full system metrics info at:')
    full_system_metrics_info_desc.add_child(f'{WEB_HOST}/{context.project.workspace.name}/{context.project.name}/experiments/{experiment.number}/system-metrics')
    system_metrics_desc.add_child(full_system_metrics_info_desc)
    root.add_child(system_metrics_desc)
    root.add_child(f'Start Command: {experiment.start_command}',
                   f'Source Code Link: {experiment.source_code_link}')

    typer.echo(root.format())


@experiment_app.command()
def logs(
    experiment_number_or_name: str = typer.Argument(..., help="The unique experiment number or name"),
    tail: int = typer.Option(200, "--tail"),
    detail: bool = typer.Option(False, "--detail", hidden=True),
    all: bool = typer.Option(False, "--all", hidden=True),
):
    """
    Display the last fifty lines of the experiment logs
    """
    context = Context(user_required=True, project_required=True)
    client = context.authorized_client

    params = {}
    if not all:
        params = {'limit': tail}
    if detail:
        params = {'withEventLog': 'true'}

    experiment_logs = client.experiment_log(
        context.project.workspace.name, context.project.name, experiment_number_or_name, params=params, raise_error=True)

    for log in experiment_logs:
        typer.echo(f'{parse_timestamp_or_none(log.timestamp)} {log.message}')

    typer.echo(
        client.get_full_info(context.project.workspace.name, context.project.name, experiment_number_or_name, 'logs')
    )


@experiment_app.command()
def run(
    command_arg: str = typer.Option(None, "--start-command", help="Start command"),
    image_arg: str = typer.Option(None, "--image", "-i", help="Kernel docker image URL"),
    cluster_name: str = typer.Option(None, "--cluster", "-c", help="Cluster name"),
    resource_name: str = typer.Option(None, "--resource", "-r", help="Resource name (for savvihub-managed cluster)"),
    processor_type: str = typer.Option(None, "--processor-type", help="cpu or gpu (for custom cluster)"),
    cpu_limit: float = typer.Option(0, "--cpu-limit", help="Number of vCPUs (for custom cluster)"),
    memory_limit: str = typer.Option(None, "--memory-limit", help="Memory capacity (ex: 4Gi, 500Mi)"),
    gpu_limit: int = typer.Option(0, "--gpu-limit", help="number of GPU cores (for custom cluster)"),
    gpu_type: str = typer.Option(None, "--gpu-type", help="GPU product name such as Tesla-K80 (for custom cluster)"),
    env_vars_arg: List[str] = typer.Option([], "-e", help="Environment variables"),
    dataset_mount_args: List[str] = typer.Option([], "--dataset", "-d", help="Dataset mounted path"),
    git_ref_arg: str = typer.Option(None, "--git-ref", help="Git commit SHA"),
    git_diff_arg: str = typer.Option(None, "--git-diff", help="Git diff file URL"),
    ignore_git_diff_arg: bool = typer.Option(False, "--ignore-git-diff", help="Ignore git diff flag"),
    project: str = typer.Option(None, "--project", hidden=True),
):
    """
    Run an experiment in SavviHub
    """
    # TODO: refactoring...
    git_ref = None
    if project:
        if not git_ref_arg:
            typer.echo("You must specify git commit SHA with `--git-ref` option")
            sys.exit(1)

        git_ref = git_ref_arg
        workspace_name, project_name = project.split("/")
        context = Context(user_required=True)

        # set project_config of context
        config_file_path = os.path.join(pathlib.Path().absolute(), ".savvihub", "config")
        if os.path.exists(config_file_path):
            remove_file(config_file_path)
        context.project_config = context.load_project_config(config_file_path)
        context.project_config.set_savvihub(url=API_HOST, workspace=workspace_name, project=project_name)
        context.project_config.save()

        # set project
        context.project = context.get_project()
    else:
        context = Context(user_required=True, project_required=True)

    client = context.authorized_client

    def find_from_args(options, selector, error_message=''):
        for option in options:
            if selector(option):
                return option

        if error_message:
            typer.echo(error_message)
            sys.exit(1)

    def find_from_inquirer(options, display, message):
        answers = inquirer.prompt([inquirer.List(
            "question",
            message=message,
            choices=[f'[{i+1}] {display(option)}' for i, option in enumerate(options)],
        )])
        answer = int(re.findall(r"[\d+]", answers.get("question"))[0]) - 1
        return options[answer]

    clusters = [c for c in client.cluster_list(context.project.workspace.name) if c.status == 'connected']
    if cluster_name:
        selected_cluster = find_from_args(
            clusters,
            lambda x: x.name == cluster_name.strip(),
            error_message=f'Cannot find cluster {cluster_name}.',
        )
    else:
        selected_cluster = find_from_inquirer(
            clusters,
            lambda x: f'{x.name}{" (SavviHub)" if x.is_savvihub_managed else f" ({x.kubernetes_master_endpoint})"}',
            "Please choose a cluster"
        )

    resource_spec_id = None
    resource_spec = None
    if selected_cluster.is_savvihub_managed:
        resources = client.kernel_resource_list(context.project.workspace.name)
        if resource_name:
            selected_resource = find_from_args(
                resources,
                lambda x: x.name == resource_name.strip(),
                error_message=f'Cannot find resource {resource_name}.',
            )
        else:
            selected_resource = find_from_inquirer(
                resources,
                lambda x: f'{x.name} ({x.description})',
                "Please choose a resource"
            )
        processor_type = selected_resource.processor_type
        resource_spec_id = selected_resource.id
    else:
        if processor_type not in ['cpu', 'gpu']:
            processor_type = find_from_inquirer(
                ['cpu', 'gpu'],
                lambda x: x,
                "Please choose a processor type",
            )
        if cpu_limit <= 0:
            cpu_limit = int(inquirer.prompt([inquirer.Text(
                INQUIRER_NAME_CPU_LIMIT,
                message="CPU limit (the number of vCPUs)",
                default=1,
            )]).get(INQUIRER_NAME_CPU_LIMIT))
            if cpu_limit <= 0:
                typer.echo('Must be greater than 0')
                sys.exit(1)
        if memory_limit is None:
            memory_limit_int = int(inquirer.prompt([inquirer.Text(
                INQUIRER_NAME_MEMORY_LIMIT,
                message="Memory limit in GiB",
                default=4,
            )]).get(INQUIRER_NAME_MEMORY_LIMIT))
            if memory_limit_int <= 0:
                typer.echo('Must be greater than 0')
                sys.exit(1)
            memory_limit = f'{memory_limit_int}Gi'
        if processor_type == 'gpu' and gpu_type is None:
            gpu_type = inquirer.prompt([inquirer.Text(  # TODO: use find_from_inquirer
                INQUIRER_NAME_GPU_LIMIT,
                message="GPU type",
                default="Tesla-K80",
            )]).get(INQUIRER_NAME_GPU_LIMIT)
        if gpu_type is not None and gpu_type != 'Empty' and gpu_limit <= 0:
            gpu_limit = int(inquirer.prompt([inquirer.Text(
                INQUIRER_NAME_GPU_LIMIT,
                message="GPU limit (the number of GPUs)",
                default=1,
            )]).get(INQUIRER_NAME_GPU_LIMIT))
            if gpu_limit <= 0:
                typer.echo('Must be greater than 0')
                sys.exit(1)

        resource_spec = {
            'processor_type': processor_type,
            'cpu_type': 'Any',
            'cpu_limit': cpu_limit,
            'memory_limit': memory_limit,
            'gpu_type': gpu_type or 'Empty',
            'gpu_limit': 0 if gpu_type is None else gpu_limit,
        }

    images = client.kernel_image_list(context.project.workspace.name, processor_type)
    if not image_arg:
        selected_image_url = find_from_inquirer(
            images,
            lambda x: f'{x.image_url} ({x.name})',
            "Please choose a kernel image"
        ).image_url
    else:
        selected_image_url = image_arg.strip()

    if command_arg:
        start_command = command_arg
    else:
        start_command = inquirer.prompt([inquirer.Text(
            INQUIRER_NAME_COMMAND,
            message="Start command",
            default="python main.py",
        )]).get(INQUIRER_NAME_COMMAND)

    dataset_mounts_parsed = []
    for dataset_mount in dataset_mount_args:
        # parse dataset and volume
        if ':' not in dataset_mount:
            typer.echo(f'Invalid dataset name: {dataset_mount}. '
                       f'You should specify dataset and mount location.\n'
                       f'ex) savvihub/mnist:3d1e0f:/input/dataset1')
            sys.exit(1)

        # parse snapshot ref
        splitted = dataset_mount.split(':')
        if len(splitted) == 2:
            dataset_name_with_workspace, mount_path = splitted
            snapshot_ref = 'latest'
        elif len(splitted) == 3:
            dataset_name_with_workspace, snapshot_ref, mount_path = splitted
        else:
            typer.echo(f'Invalid dataset name: {dataset_mount}\n'
                       f'You should specify dataset and mount location.\n'
                       f'ex) savvihub/mnist:3d1e0f:/input/dataset1')
            sys.exit(1)

        if '/' not in dataset_name_with_workspace:
            typer.echo(f'Invalid dataset name: {dataset_mount}\n'
                       f'You should specify dataset with workspace.\n'
                       f'ex) savvihub/mnist:3d1e0f:/input/dataset1')
            sys.exit(1)

        workspace_name, dataset_name = dataset_name_with_workspace.split('/')

        # read dataset or snapshot
        dataset_obj = client.dataset_read(workspace_name, dataset_name)
        if not dataset_obj:
            typer.echo(f'Invalid dataset: {dataset_name_with_workspace}\n'
                       f'Please check your dataset exist in savvihub.\n')
            sys.exit(1)

        if snapshot_ref != 'latest':
            snapshot_obj = client.snapshot_read(dataset_obj.volume_id, snapshot_ref)
            if not snapshot_obj:
                typer.echo(f'Invalid dataset snapshots: {dataset_mount}\n'
                           f'Please check your dataset and snapshot exist in savvihub.')
                sys.exit(1)

        if not mount_path.startswith('/input'):
            typer.echo(f'Invalid dataset mount path: {mount_path}\n'
                       f'Dataset mount path should start with /input.')
            sys.exit(1)

        dataset_mounts_parsed.append(dict(
            dataset_id=dataset_obj.id,
            snapshot_ref=snapshot_ref,
            mount_path=mount_path,
        ))

    diff_file = None
    if git_ref:
        pass
    elif git_ref_arg:
        if not context.git_repo.check_revision_in_remote(git_ref_arg.strip()):
            typer.echo(f'Git commit {git_ref_arg.strip()} does not exist in a remote repository.')
            sys.exit(1)

        git_ref = git_ref_arg.strip()

        if git_diff_arg and (git_diff_arg.startswith('https://') or git_diff_arg.startswith('http://')):
            diff_file = tempfile.NamedTemporaryFile(suffix='.patch')
            d = DownloadableFileObject(git_diff_arg, os.path.dirname(diff_file.name), os.path.basename(diff_file.name))
            d.download(session=context.authorized_client.session)
            diff_file.seek(0)

    else:
        git_ref, branch, is_head = context.git_repo.get_remote_revision_or_branch()
        commit = context.git_repo.get_commit_message(git_ref)
        typer.echo(f'Run experiment with revision {git_ref[:6]} ({branch})')
        typer.echo(f'Commit: {commit}')
        if not is_head:
            typer.echo('Your current revision does not exist in remote repository. SavviHub will use latest remote '
                       'branch revision hash and diff.')
        typer.echo('')

        has_diff, diff_status = context.git_repo.get_current_diff_status(git_ref)
        if has_diff and not ignore_git_diff_arg:
            typer.echo('Diff to be uploaded: ')

            uncommitted_files = diff_status.get('uncommitted')
            untracked_files = diff_status.get('untracked')

            if uncommitted_files:
                typer.echo('  Changes not committed')
                typer.echo('\n'.join([f'    {x}' for x in uncommitted_files]))
                typer.echo('')
            if untracked_files:
                typer.echo(f'  Untracked files:')
                typer.echo('\n'.join([f'    {x}' for x in untracked_files]))
                typer.echo('')

            # TODO: reduce choices if uncommitted or untracked not exist

            answers = inquirer.prompt([inquirer.List(
                'experiment',
                message='Run experiment with diff?',
                choices=[
                    '[1] Run experiment with uncommitted and untracked changes.',
                    '[2] Run experiment with uncommitted changes.',
                    '[3] Run experiment without any changes.',
                    '[4] Abort.',
                ],
            )])
            answer = int(re.findall(r'[\d+]', answers.get('experiment'))[0])

            diff_file = None
            if answer == 1:
                diff_file = context.git_repo.get_current_diff_file(git_ref, with_untracked=True)
            elif answer == 2:
                diff_file = context.git_repo.get_current_diff_file(git_ref, with_untracked=False)
            elif answer == 3:
                pass
            else:
                typer.echo('Aborted.')
                return

    diff_file_path = None
    if diff_file:
        typer.echo('Generating diff patch file...')
        uploaded = Uploader.upload(
            context,
            local_path=diff_file.name,
            volume_id=context.project.volume_id,
            remote_path=os.path.basename(diff_file.name),
            progressable=typer.progressbar,
        )
        diff_file_path = uploaded.path
        diff_file.close()

    env_vars = []
    for env_var in env_vars_arg:
        try:
            env_key, env_value = env_var.split("=")
            env_vars.append({
                'key': env_key,
                'value': env_value,
            })
        except ValueError:
            typer.echo(f'Cannot parse environment variable: {env_var}')
            sys.exit(1)

    res = client.experiment_create(
        workspace=context.project.workspace.name,
        project=context.project.name,
        cluster_name=selected_cluster.name,
        image_url=selected_image_url,
        resource_spec=resource_spec,
        resource_spec_id=resource_spec_id,
        git_ref=git_ref,
        git_diff_file_path=diff_file_path,
        start_command=start_command,
        datasets=dataset_mounts_parsed,
        env_vars=env_vars,
    )

    res_data = res.json()
    if res.status_code == 400:
        typer.echo(get_error_message(res_data))
        sys.exit(1)

    res.raise_for_status()

    experiment_number = res_data.get('number')
    typer.echo(f'Experiment {experiment_number} is running. Check the experiment status at below link')
    typer.echo(f'{WEB_HOST}/{context.project.workspace.name}/{context.project.name}/experiments/{experiment_number}')


@experiment_output_app.callback()
def output_main():
    """
    Manage experiment output files
    """


@experiment_output_app.command()
def ls(
    experiment_number_or_name: str = typer.Argument(..., help="The unique experiment number or name"),
    path: str = typer.Argument(None, help='Output file path'),
    recursive: bool = typer.Option(False, '-r', '--recursive', help='recursive flag'),
    directory: bool = typer.Option(False, '-d', '--directory',
                                   help='list the directory itself, not its contents'),
):
    """
    List the output files of the experiment
    """
    context = Context(user_required=True, project_required=True)
    client = context.authorized_client

    experiment = client.experiment_read(
        context.project.workspace.name, context.project.name, experiment_number_or_name, raise_error=True,
    )
    volume_file_list(experiment.output_volume.volume_id, 'latest', path or '', recursive, directory)


@experiment_output_app.command()
def download(
    experiment_number_or_name: str = typer.Argument(..., help="The unique experiment number or name"),
    dest_path: str = typer.Argument(None, help='The files will be downloaded to ./output if omitted.'),
):
    """
    Download experiment output files
    """
    context = Context(user_required=True, project_required=True)
    client = context.authorized_client

    experiment = client.experiment_read(
        context.project.workspace.name, context.project.name, experiment_number_or_name, raise_error=True,
    )
    volume_file_copy(experiment.output_volume.volume_id, 'latest', '',
                     None, None, dest_path or './output',
                     recursive=True, watch=False)


if __name__ == "__main__":
    experiment_app()



# coding: utf-8

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


import inspect
import pprint
import re  # noqa: F401
import six

from savvihub_client.configuration import Configuration


class ModelResourceRequirements(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'guarantee': 'int',
        'types': 'list[str]',
        'version': 'int'
    }

    attribute_map = {
        'guarantee': 'guarantee',
        'types': 'types',
        'version': 'version'
    }

    def __init__(self, guarantee=None, types=None, version=None, local_vars_configuration=None):  # noqa: E501
        """ModelResourceRequirements - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration.get_default_copy()
        self.local_vars_configuration = local_vars_configuration

        self._guarantee = None
        self._types = None
        self._version = None
        self.discriminator = None

        if guarantee is not None:
            self.guarantee = guarantee
        if types is not None:
            self.types = types
        if version is not None:
            self.version = version

    @property
    def guarantee(self):
        """Gets the guarantee of this ModelResourceRequirements.  # noqa: E501


        :return: The guarantee of this ModelResourceRequirements.  # noqa: E501
        :rtype: int
        """
        return self._guarantee

    @guarantee.setter
    def guarantee(self, guarantee):
        """Sets the guarantee of this ModelResourceRequirements.


        :param guarantee: The guarantee of this ModelResourceRequirements.  # noqa: E501
        :type guarantee: int
        """

        self._guarantee = guarantee

    @property
    def types(self):
        """Gets the types of this ModelResourceRequirements.  # noqa: E501


        :return: The types of this ModelResourceRequirements.  # noqa: E501
        :rtype: list[str]
        """
        return self._types

    @types.setter
    def types(self, types):
        """Sets the types of this ModelResourceRequirements.


        :param types: The types of this ModelResourceRequirements.  # noqa: E501
        :type types: list[str]
        """

        self._types = types

    @property
    def version(self):
        """Gets the version of this ModelResourceRequirements.  # noqa: E501


        :return: The version of this ModelResourceRequirements.  # noqa: E501
        :rtype: int
        """
        return self._version

    @version.setter
    def version(self, version):
        """Sets the version of this ModelResourceRequirements.


        :param version: The version of this ModelResourceRequirements.  # noqa: E501
        :type version: int
        """

        self._version = version

    def to_dict(self, serialize=False):
        """Returns the model properties as a dict"""
        result = {}

        def convert(x):
            if hasattr(x, "to_dict"):
                args = inspect.getargspec(x.to_dict).args
                if len(args) == 1:
                    return x.to_dict()
                else:
                    return x.to_dict(serialize)
            else:
                return x

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            attr = self.attribute_map.get(attr, attr) if serialize else attr
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: convert(x),
                    value
                ))
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], convert(item[1])),
                    value.items()
                ))
            else:
                result[attr] = convert(value)

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ModelResourceRequirements):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ModelResourceRequirements):
            return True

        return self.to_dict() != other.to_dict()



from savvihub.savvihub_cli.main import __version__
from tests.conftest import BaseTestCase


class BasicTest(BaseTestCase):
    def test_version(self):
        result = self.runner.invoke(['--version'])
        assert f'SavviHub CLI Version: {__version__}' in result.stdout

    def test_ping(self):
        result = self.runner.invoke(['ping'])
        assert 'Response code: 200' in result.stdout
        assert 'Response text: pong' in result.stdout



from savvihub.common.context import Context


global_context = None


def log(step, row=None):
    """
    step: a step for each iteration (required)
    row: a dictionary to log
    """
    global global_context
    if global_context is None:
        global_context = Context(experiment_required=True)
    global_context.experiment.log(row=row, step=step)



# coding: utf-8

# flake8: noqa

"""
    Aron API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 1.0.0
    Generated by: https://openapi-generator.tech
"""


from __future__ import absolute_import

__version__ = "1.0.0"

# import apis into sdk package
from savvihub_client.api.api_v1_api import APIV1Api

# import ApiClient
from savvihub_client.api_client import ApiClient
from savvihub_client.configuration import Configuration
from savvihub_client.exceptions import OpenApiException
from savvihub_client.exceptions import ApiTypeError
from savvihub_client.exceptions import ApiValueError
from savvihub_client.exceptions import ApiKeyError
from savvihub_client.exceptions import ApiAttributeError
from savvihub_client.exceptions import ApiException
# import models into sdk package
from savvihub_client.models.access_token_create_api import AccessTokenCreateAPI
from savvihub_client.models.account_invitation_token_validate_response import AccountInvitationTokenValidateResponse
from savvihub_client.models.account_sign_in_cli_check_response import AccountSignInCliCheckResponse
from savvihub_client.models.account_sign_in_cli_token_response import AccountSignInCliTokenResponse
from savvihub_client.models.change_password_api import ChangePasswordAPI
from savvihub_client.models.change_username_api import ChangeUsernameAPI
from savvihub_client.models.cli_experiment_metrics_update_api import CliExperimentMetricsUpdateAPI
from savvihub_client.models.cluster_all_cluster_list_response import ClusterAllClusterListResponse
from savvihub_client.models.cluster_custom_cluster_list_response import ClusterCustomClusterListResponse
from savvihub_client.models.custom_cluster_add_api import CustomClusterAddAPI
from savvihub_client.models.custom_cluster_update_api import CustomClusterUpdateAPI
from savvihub_client.models.dataset_update_api import DatasetUpdateAPI
from savvihub_client.models.experiment_create_api import ExperimentCreateAPI
from savvihub_client.models.experiment_experiment_list_response import ExperimentExperimentListResponse
from savvihub_client.models.experiment_experiment_log_response import ExperimentExperimentLogResponse
from savvihub_client.models.experiment_experiment_metrics_response import ExperimentExperimentMetricsResponse
from savvihub_client.models.experiment_experiment_parameter_list_response import ExperimentExperimentParameterListResponse
from savvihub_client.models.experiment_experiment_system_metrics_response import ExperimentExperimentSystemMetricsResponse
from savvihub_client.models.experiment_update_api import ExperimentUpdateAPI
from savvihub_client.models.external_git_hub_config_response import ExternalGitHubConfigResponse
from savvihub_client.models.external_git_hub_repository_list_response import ExternalGitHubRepositoryListResponse
from savvihub_client.models.external_google_config_response import ExternalGoogleConfigResponse
from savvihub_client.models.external_slack_config_response import ExternalSlackConfigResponse
from savvihub_client.models.gs_dataset_create_api import GSDatasetCreateAPI
from savvihub_client.models.git_hub_authorize_api import GitHubAuthorizeAPI
from savvihub_client.models.github_code_of_conduct import GithubCodeOfConduct
from savvihub_client.models.github_license import GithubLicense
from savvihub_client.models.github_match import GithubMatch
from savvihub_client.models.github_organization import GithubOrganization
from savvihub_client.models.github_plan import GithubPlan
from savvihub_client.models.github_repository import GithubRepository
from savvihub_client.models.github_text_match import GithubTextMatch
from savvihub_client.models.github_user import GithubUser
from savvihub_client.models.invitation_token_validate_api import InvitationTokenValidateAPI
from savvihub_client.models.kernel_kernel_image_list_response import KernelKernelImageListResponse
from savvihub_client.models.kernel_kernel_resource_spec_list_response import KernelKernelResourceSpecListResponse
from savvihub_client.models.model_access_token import ModelAccessToken
from savvihub_client.models.model_env_var import ModelEnvVar
from savvihub_client.models.model_experiment_metrics_summary import ModelExperimentMetricsSummary
from savvihub_client.models.model_field_property import ModelFieldProperty
from savvihub_client.models.model_file_metadata import ModelFileMetadata
from savvihub_client.models.model_kernel_cluster import ModelKernelCluster
from savvihub_client.models.model_kernel_cluster_config import ModelKernelClusterConfig
from savvihub_client.models.model_kernel_cluster_config_ingress import ModelKernelClusterConfigIngress
from savvihub_client.models.model_kernel_cluster_config_storage_class import ModelKernelClusterConfigStorageClass
from savvihub_client.models.model_kernel_cluster_quota import ModelKernelClusterQuota
from savvihub_client.models.model_kernel_image import ModelKernelImage
from savvihub_client.models.model_kernel_resource_spec import ModelKernelResourceSpec
from savvihub_client.models.model_permission_data import ModelPermissionData
from savvihub_client.models.model_permission_detail import ModelPermissionDetail
from savvihub_client.models.model_permissions import ModelPermissions
from savvihub_client.models.model_resource_requirements import ModelResourceRequirements
from savvihub_client.models.model_resource_spec_fields import ModelResourceSpecFields
from savvihub_client.models.model_service_config import ModelServiceConfig
from savvihub_client.models.model_service_config_jupyter_lab import ModelServiceConfigJupyterLab
from savvihub_client.models.model_service_config_tensorboard import ModelServiceConfigTensorboard
from savvihub_client.models.model_service_endpoint import ModelServiceEndpoint
from savvihub_client.models.model_service_port import ModelServicePort
from savvihub_client.models.model_service_ports import ModelServicePorts
from savvihub_client.models.model_service_status import ModelServiceStatus
from savvihub_client.models.model_service_status_jupyter_lab import ModelServiceStatusJupyterLab
from savvihub_client.models.model_service_status_tensorboard import ModelServiceStatusTensorboard
from savvihub_client.models.model_user import ModelUser
from savvihub_client.models.model_volume_source_dataset import ModelVolumeSourceDataset
from savvihub_client.models.model_volume_source_project import ModelVolumeSourceProject
from savvihub_client.models.model_volume_source_snapshot import ModelVolumeSourceSnapshot
from savvihub_client.models.model_volume_source_volume import ModelVolumeSourceVolume
from savvihub_client.models.model_workspace import ModelWorkspace
from savvihub_client.models.project_create_api import ProjectCreateAPI
from savvihub_client.models.project_git_hub_create_api import ProjectGitHubCreateAPI
from savvihub_client.models.project_project_branch_list_response import ProjectProjectBranchListResponse
from savvihub_client.models.project_project_commit_list_response import ProjectProjectCommitListResponse
from savvihub_client.models.project_update_api import ProjectUpdateAPI
from savvihub_client.models.proto_branch import ProtoBranch
from savvihub_client.models.proto_commit import ProtoCommit
from savvihub_client.models.proto_commit_author import ProtoCommitAuthor
from savvihub_client.models.proto_container_system_metrics import ProtoContainerSystemMetrics
from savvihub_client.models.proto_current_system_metric import ProtoCurrentSystemMetric
from savvihub_client.models.proto_metric import ProtoMetric
from savvihub_client.models.proto_node_system_metrics import ProtoNodeSystemMetrics
from savvihub_client.models.proto_system_metric import ProtoSystemMetric
from savvihub_client.models.proto_volume_mount_request import ProtoVolumeMountRequest
from savvihub_client.models.proto_volume_mount_request_source_dataset import ProtoVolumeMountRequestSourceDataset
from savvihub_client.models.proto_volume_mount_request_source_output import ProtoVolumeMountRequestSourceOutput
from savvihub_client.models.proto_volume_mount_request_source_project import ProtoVolumeMountRequestSourceProject
from savvihub_client.models.proto_volume_mount_request_source_snapshot import ProtoVolumeMountRequestSourceSnapshot
from savvihub_client.models.proto_volume_mount_request_source_volume import ProtoVolumeMountRequestSourceVolume
from savvihub_client.models.proto_volume_mount_request_source_working_dir import ProtoVolumeMountRequestSourceWorkingDir
from savvihub_client.models.proto_workload_log import ProtoWorkloadLog
from savvihub_client.models.redis_experiment_metric import RedisExperimentMetric
from savvihub_client.models.response_dataset import ResponseDataset
from savvihub_client.models.response_dataset_list import ResponseDatasetList
from savvihub_client.models.response_dataset_source import ResponseDatasetSource
from savvihub_client.models.response_experiment import ResponseExperiment
from savvihub_client.models.response_experiment_history import ResponseExperimentHistory
from savvihub_client.models.response_experiment_kernel_cluster import ResponseExperimentKernelCluster
from savvihub_client.models.response_file_metadata import ResponseFileMetadata
from savvihub_client.models.response_kernel_cluster import ResponseKernelCluster
from savvihub_client.models.response_kernel_cluster_node import ResponseKernelClusterNode
from savvihub_client.models.response_kernel_cluster_node_gpu import ResponseKernelClusterNodeGPU
from savvihub_client.models.response_kernel_cluster_node_info import ResponseKernelClusterNodeInfo
from savvihub_client.models.response_kernel_cluster_system_metrics import ResponseKernelClusterSystemMetrics
from savvihub_client.models.response_kernel_resource_spec import ResponseKernelResourceSpec
from savvihub_client.models.response_my_user import ResponseMyUser
from savvihub_client.models.response_project import ResponseProject
from savvihub_client.models.response_project_experiment_summary import ResponseProjectExperimentSummary
from savvihub_client.models.response_project_link import ResponseProjectLink
from savvihub_client.models.response_project_list_response import ResponseProjectListResponse
from savvihub_client.models.response_sample_project import ResponseSampleProject
from savvihub_client.models.response_sample_project_list_response import ResponseSampleProjectListResponse
from savvihub_client.models.response_service import ResponseService
from savvihub_client.models.response_service_template import ResponseServiceTemplate
from savvihub_client.models.response_snapshot import ResponseSnapshot
from savvihub_client.models.response_user import ResponseUser
from savvihub_client.models.response_user_with_token_response import ResponseUserWithTokenResponse
from savvihub_client.models.response_volume import ResponseVolume
from savvihub_client.models.response_volume_dataset import ResponseVolumeDataset
from savvihub_client.models.response_volume_experiment import ResponseVolumeExperiment
from savvihub_client.models.response_volume_mount_info import ResponseVolumeMountInfo
from savvihub_client.models.response_volume_source_dataset import ResponseVolumeSourceDataset
from savvihub_client.models.response_volume_source_project import ResponseVolumeSourceProject
from savvihub_client.models.response_volume_source_snapshot import ResponseVolumeSourceSnapshot
from savvihub_client.models.response_volume_source_volume import ResponseVolumeSourceVolume
from savvihub_client.models.response_workspace import ResponseWorkspace
from savvihub_client.models.response_workspace_member import ResponseWorkspaceMember
from savvihub_client.models.response_workspace_notification_config import ResponseWorkspaceNotificationConfig
from savvihub_client.models.s3_dataset_create_api import S3DatasetCreateAPI
from savvihub_client.models.savvi_hub_dataset_create_api import SavviHubDatasetCreateAPI
from savvihub_client.models.service_config_update_api import ServiceConfigUpdateAPI
from savvihub_client.models.service_create_api import ServiceCreateAPI
from savvihub_client.models.service_name_update_api import ServiceNameUpdateAPI
from savvihub_client.models.service_resource_spec_update_api import ServiceResourceSpecUpdateAPI
from savvihub_client.models.service_service_list_response import ServiceServiceListResponse
from savvihub_client.models.service_service_log_response import ServiceServiceLogResponse
from savvihub_client.models.service_service_system_metrics_response import ServiceServiceSystemMetricsResponse
from savvihub_client.models.service_service_template_list_response import ServiceServiceTemplateListResponse
from savvihub_client.models.sign_in_api import SignInAPI
from savvihub_client.models.sign_in_cli_confirm_api import SignInCliConfirmAPI
from savvihub_client.models.sign_in_git_hub_api import SignInGitHubAPI
from savvihub_client.models.sign_in_google_api import SignInGoogleAPI
from savvihub_client.models.sign_up_api import SignUpAPI
from savvihub_client.models.sign_up_validate_api import SignUpValidateAPI
from savvihub_client.models.slack_authorize_api import SlackAuthorizeAPI
from savvihub_client.models.snapshot_create_api import SnapshotCreateAPI
from savvihub_client.models.snapshot_snapshot_list_response import SnapshotSnapshotListResponse
from savvihub_client.models.storage_file_action_url_info import StorageFileActionURLInfo
from savvihub_client.models.types_null_json_text import TypesNullJSONText
from savvihub_client.models.verify_email_api import VerifyEmailAPI
from savvihub_client.models.volume_file_copy_api import VolumeFileCopyAPI
from savvihub_client.models.volume_file_create_api import VolumeFileCreateAPI
from savvihub_client.models.volume_volume_file_copy_response import VolumeVolumeFileCopyResponse
from savvihub_client.models.volume_volume_file_delete_response import VolumeVolumeFileDeleteResponse
from savvihub_client.models.volume_volume_file_list_response import VolumeVolumeFileListResponse
from savvihub_client.models.workspace_create_api import WorkspaceCreateAPI
from savvihub_client.models.workspace_member_add_api import WorkspaceMemberAddAPI
from savvihub_client.models.workspace_notification_config_email_add_api import WorkspaceNotificationConfigEmailAddAPI
from savvihub_client.models.workspace_update_api import WorkspaceUpdateAPI
from savvihub_client.models.workspace_workspace_list_response import WorkspaceWorkspaceListResponse
from savvihub_client.models.workspace_workspace_member_list_response import WorkspaceWorkspaceMemberListResponse




import os
import sys
from urllib.parse import urlparse

import typer

from savvihub import Context
from savvihub.api.savvihub import SavviHubClient
from savvihub.api.types import SavviHubDataset
from savvihub.savvihub_cli.volume import volume_file_list, volume_file_remove, volume_file_copy

dataset_files_app = typer.Typer()


class PathException(Exception):
    pass


def refine_dataset_path(
    path_arg,
    raise_if_not_empty=False,
    raise_if_not_dir=False,
    raise_if_not_exist=False,
    remote_only=False,
):
    if path_arg.startswith('sv://'):
        typer.echo('[Error] Use \'svds://\' for datasets.')
        sys.exit(1)

    if path_arg.startswith('svds://') or remote_only:
        path_arg = path_arg.replace('svds://', '')
        if path_arg.strip('/').count('/') == 0:
            typer.echo('[Error] Dataset not found!\n You should specify dataset with workspace. ex) savvihub/mnist')
            sys.exit(1)
        return 'svds://' + path_arg, True

    path = os.path.abspath(path_arg)
    if path_arg.endswith('/'):
        path += '/'

    if os.path.exists(path):
        if not os.path.isdir(path) and raise_if_not_dir:
            raise PathException(f'Must specify directory: {path_arg}')
        if raise_if_not_empty and len(os.listdir(path)) > 0:
            raise PathException(f'Must specify empty directory: {path_arg}')
    else:
        if raise_if_not_exist:
            raise PathException(f'Must specify directory: {path_arg}')

    return path, False


def parse_remote_dataset_path(remote_path: str) -> (SavviHubDataset, str, str):
    u = urlparse(remote_path)
    workspace = u.netloc
    path = u.path.lstrip('/')
    split_path = path.split('/', 1)
    split_snapshot = split_path[0].split(':', 1)

    file_path = split_path[1] if len(split_path) == 2 else ''
    snapshot = split_snapshot[1] if len(split_snapshot) == 2 else 'latest'

    return get_dataset(workspace, split_snapshot[0]), snapshot, file_path


def get_dataset(workspace: str, dataset: str) -> SavviHubDataset:
    context = Context(user_required=True)
    client = context.authorized_client

    r = client.dataset_read(workspace, dataset)
    if not r:
        typer.echo('[Error] Dataset not found!\n You should specify dataset with workspace. ex) savvihub/mnist')
        sys.exit(1)
    return r


@dataset_files_app.callback()
def main():
    """
    Manage files in the dataset
    """
    return


@dataset_files_app.command()
def ls(
    source_path_arg: str = typer.Argument(...),
    recursive: bool = typer.Option(False, '-r', '--recursive', help='recursive flag'),
    directory: bool = typer.Option(False, '-d', '--directory', help='list the directory itself, not its contents')
):
    """
    List files in the dataset with prefix
    """
    try:
        source_path, is_source_remote = refine_dataset_path(source_path_arg, raise_if_not_exist=True, remote_only=True)
    except PathException as e:
        typer.echo(str(e))
        return

    dataset, snapshot, path = parse_remote_dataset_path(source_path)
    volume_file_list(dataset.volume_id, snapshot, path, recursive, directory)


@dataset_files_app.command()
def rm(
    source_path_arg: str = typer.Argument(...),
    recursive: bool = typer.Option(False, '-r', '-R', '--recursive',
                                   help='Remove directories and their contents recursively'),
):
    """
    Remove files in the volume with path
    """
    try:
        source_path, is_source_remote = refine_dataset_path(source_path_arg, raise_if_not_exist=True, remote_only=True)
    except PathException as e:
        typer.echo(str(e))
        return

    dataset, snapshot, path = parse_remote_dataset_path(source_path)
    volume_file_remove(dataset.volume_id, snapshot, path, recursive)


@dataset_files_app.command()
def cp(
    source_path_arg: str = typer.Argument(...),
    dest_path_arg: str = typer.Argument(...),
    recursive: bool = typer.Option(False, '-r', '--recursive'),
    watch: bool = typer.Option(False, '-w', '--watch'),
):
    try:
        source_path, is_source_remote = refine_dataset_path(source_path_arg, raise_if_not_exist=True)
        dest_path, is_dest_remote = refine_dataset_path(dest_path_arg)
    except PathException as e:
        typer.echo(str(e))
        return

    source_volume_id = dest_volume_id = None
    source_snapshot = dest_snapshot = None
    if is_source_remote:
        source_dataset, source_snapshot, source_path = parse_remote_dataset_path(source_path)
        source_volume_id = source_dataset.volume_id
    if is_dest_remote:
        dest_dataset, dest_snapshot, dest_path = parse_remote_dataset_path(dest_path)
        dest_volume_id = dest_dataset.volume_id

    volume_file_copy(source_volume_id, source_snapshot, source_path,
                     dest_volume_id, dest_snapshot, dest_path,
                     recursive, watch)
